// this file contains: Application startup, run, & shutdown code
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 
// To do:
// 
// - in ParseArguments check use strncpy or report error to avoid 
//   segment violations on long parameters
// 

#include "sqBeApp.h"
#include "sqBeGlobal.h"
#include "sqBeCopyBits.h"
#include "sqBeExternalPrims.h"
#include "sqBeIOPrims.h"
#include "sqBeView.h"
#include "sqBeWindow.h"
#include "sqBeDirectWindow.h"
#include <Alert.h>
#include <Directory.h>
#include <FindDirectory.h>
#include <File.h>
#include <Entry.h>
#include <string.h>
#include <ctype.h>

void Cleanup();                                  // any cleanup before shutdown
void InitSqueakToBeColorMap(BScreen *screen);    // creates a Squeak to Be color conversion table
extern "C" int ioInitSecurity(void); // implemented in sqBeSecurity.c
void ParseArguments(int argc, char **argv);      // parses the command-line arguments
void ShowAlert(const char *err);                 // shows a BAlert to the user with the given error
void ShowUsage();                                // send description of command line options to screen
int32 StartInterpret(void * data);                // starts the Squeak interpreter

// variables used by the BeOS-specific part of the VM
const int32 DefaultMemorySize = 32 << 20; // default for MemorySize (below)
bool ForceBWindow = false;               // indicates if command line params
                                              // have forced the use of a BWindow 
bool ForceDirectWindow = false;          // indicates if command line params
                                              // have forced the use of a BDirectWindow
char imageName[B_FILE_NAME_LENGTH];      // full path to image
thread_id InterpretThread;                    // thread for interpreter  
volatile int low_res_time;	// a variable that is written to by the view's Pulse() method to become a low-resolution timer
int32 MemorySize = DefaultMemorySize;     // amount of memory to allocate for image
int *Pubargc;                            // public access argc (used by GetAttribute)
char ***Pubargv;                            // public access argv (used by GetAttribute)
int PubImageArg;                        // number of image argument (used by GetAttribute)
int ProfCount[MAX_PROFILE_POINTS] = PROFILE_INIT_M1;
bigtime_t ProfElapsed[MAX_PROFILE_POINTS] = PROFILE_INIT_M1;
bigtime_t ProfMax[MAX_PROFILE_POINTS] = PROFILE_INIT_M1;
bigtime_t ProfMin[MAX_PROFILE_POINTS] = PROFILE_INIT_M1;
const char *ProfNames[MAX_PROFILE_POINTS] = PROFILE_NAMES;
char *ProgName;                           // name of the VM file
volatile bool Quitting = false;                   // is Squeak in the process of quiting?
char ShortImageName[B_FILE_NAME_LENGTH]; // just the image file name
port_id SqSem = 0; // port for asynchronous signalling of Squeak semaphores
port_id SqEvents = 0; // port for sending Squeak events to, if non-zero indicates event-based image
int SqEventSem; // semaphore for signalling squeak events
char VMPath[B_PATH_NAME_LENGTH];         // full path to interpreter's directory

// pointers to the single instances of the squeak window/app classes
SqueakApp *SqApp = 0;
BWindow *SqWindow = 0;
BView *SqView = 0;

///////////////
// SqueakApp //
///////////////

#define APP_SIG					"application/x-BeSqueak"
#define SQUEAK_IMAGE_MIMETYPE	"application/squeak-image"
#define PREF_DEFER_UPDATES "deferUpdates "

// BeSqueak's application constructor simply passes along the application's MIME signature
SqueakApp::SqueakApp() : BApplication(APP_SIG) {
}
	
// show about box when requested
void SqueakApp::AboutRequested() {
	BAlert *alert = new BAlert("About BeSqueak", 
                    "Virtual Machine:\n\tBeSqueak 1.1 - Released 1/2/2001\n\nInterpreter:\n\tSqueak v3.1alpha\n\nPorted by\n\tColin Sarsfield\n\tTrey Bordeau\n\tRichard Peskin\n\nIcon thanks to William Bull", 
                    "Squeak!", NULL, NULL, B_WIDTH_FROM_LABEL, B_IDEA_ALERT); 
	alert->Go();
}		

// parse command line arguments	
void SqueakApp::ArgvReceived(int32 argc, char **argv) {
	ParseArguments(argc, argv);
	BApplication::ArgvReceived(argc, argv);
}

// load preferences file -- right now the prefs file only stores the "Defer Updates" flag
void SqueakApp::LoadPreferences() {
	// lookup user settings directory
	if (find_directory(B_USER_SETTINGS_DIRECTORY, &PrefDirPath) ==  B_OK) {
		
		PrefDirPath.SetTo(PrefDirPath.Path(), "BeSqueak");
		GenPrefFilePath.SetTo(PrefDirPath.Path(), "Preferences");

		BFile PrefFile;
		
		// open general preferences file
		if (PrefFile.SetTo(GenPrefFilePath.Path(), B_READ_ONLY) == B_OK) {
			off_t file_size;
			char *data;

			PrefFile.GetSize(&file_size);
	
			if (file_size > 0) {
				data = new char[file_size + 1];
				if (data == NULL) {
					printf("Out of memory while loading prefs file");
					exit(-1);
				}
			
				// load the entire file into memory
				PrefFile.Seek(0, SEEK_SET);
				PrefFile.Read((void *)data, file_size);
				data[file_size] = '\0';
		
				char *cur_line = data;
				char *next_line = data;
				// parse the preferences out
				// preferences take the form:
				//   deferUpdates n
				//   deferUpdates y
				// preferences are case sensitive
				while (next_line != NULL) {
					next_line = strchr(cur_line, '\n');
					if (next_line != NULL)
						*next_line = '\0';

					if (strncmp(cur_line, PREF_DEFER_UPDATES, strlen(PREF_DEFER_UPDATES)) == 0)
						fDeferredUpdate = tolower(cur_line[strlen(PREF_DEFER_UPDATES)]) == 'y';

					cur_line = next_line + 1;
				}			
	
				delete[] data;
			}
			PrefFile.Unset();
		}
	}
}

void SqueakApp::Quit() {
	// make sure to save preferences before quitting the VM entirely
	SavePreferences();
	BApplication::Quit();
}

// startup sequence
// 1. load prefs
// 2. run squeak image
// 3. start main application looper
void SqueakApp::ReadyToRun() {
	LoadPreferences();
	RunImage();
	BApplication::ReadyToRun();
}
	
// startup squeak image
void SqueakApp::RunImage() {
	BScreen screen;
	int ret;
	
	// build image file's name, full path, and the vm's path.  For example,
	//
	// ShortImageName = 'Squeak2.6.image'
	// imageName = '/boot/home/dev/BeSqueak work/Squeak2.6.image'
	// VMPath = '/boot/home/dev/BeSqueak work/'
	BEntry entry(ShortImageName, true), vmEntry(*Pubargv[0], true);
	BPath path; 
	entry.GetPath(&path); 
	strcpy(ShortImageName, path.Leaf());
	strcpy(imageName, path.Path());
	vmEntry.GetPath(&path);
	path.GetParent(&path);
	strcpy(VMPath, path.Path());
	strcat(VMPath, "/");
		
	#if defined(DEBUGSQUEAK)
		printf("ShortImageName = '%s'\n", ShortImageName);
		printf("imageName = '%s'\n", imageName);
		printf("VMPath = '%s'\n", VMPath);
	#endif
	
	// allocate port for passing keyboard input from one thread to another
	if ((KeyBuf = create_port(128, "keyBuf_port")) < B_NO_ERROR) {
		ShowAlert("Unable to create keyboard buffer port.");	
		exit(-3);
	}
	// allocate port for passing squeak semaphore signals from one thread to another
	if ((SqSem = create_port(128, "squeak_semaphore_port")) < B_NO_ERROR) {
		ShowAlert("Unable to create squeak semaphore port.");	
		exit(-3);
	}

	// specify cleanup routine to run during program's exit
	atexit(Cleanup);
		
	// create a suspended thread for the interpreter
	InterpretThread = spawn_thread(&StartInterpret, "Interpreter", B_NORMAL_PRIORITY, 0); 
		 
	if (InterpretThread < B_NO_ERROR) {
		// failure
		ShowAlert("Unable to spawn main interpreter thread.");
		exit(-3);			
	}
			
	// build a map of squeak colors to screen colors
	InitSqueakToBeColorMap(&screen);
			
	// start the interpreter thread running
	if ((ret = resume_thread(InterpretThread)) != B_NO_ERROR) {
		printf("Main interpreter thread error: ");
		switch(ret) {
			case B_INTERRUPTED:
				printf("Interrupted\n");
				break;
			case B_BAD_THREAD_ID:
				printf("Bad Id\n");
				break;
			case B_BAD_THREAD_STATE:
				printf("Bad State\n");
				break;
			default:
				printf("Unknown (%d)\n", ret);
		}
		exit(-2);
	}
}

// this supports drag & drop of an image file onto the VM
void SqueakApp::RefsReceived(BMessage *message) {
	entry_ref 	ref;		// The entry_ref to open
	
	BApplication::RefsReceived(message);

	if (message->FindRef("refs", 0, &ref) != B_OK)
		return;

	BEntry entry(&ref); 
	BPath path;

	entry.GetPath(&path);
	strcpy(ShortImageName, path.Path());
}
	
// save preferences (see LoadPreferences for more details)	
void SqueakApp::SavePreferences() {
	// create a directory for BeSqueak preferences - for now there is just one preferences file in it
	if (create_directory(PrefDirPath.Path(), 0777) == B_OK) {
		BFile PrefFile;
		
		if (PrefFile.SetTo(GenPrefFilePath.Path(), B_WRITE_ONLY | B_CREATE_FILE) == B_OK) {
			PrefFile.Seek(0, SEEK_SET);
			char str[256];
			sprintf(str, "%s%c\n", PREF_DEFER_UPDATES, fDeferredUpdate ? 'y' : 'n');
			PrefFile.Write(str, strlen(str));

			PrefFile.SetSize(PrefFile.Position());
			PrefFile.Unset();
		}
	}
}

/////////////////////////////////
// BeOS specific suport & main //
/////////////////////////////////

// signal a squeak semaphore asynchronously.  This allows the window threads and other threads outside of the
// interpreter thread to signal the squeak interpreter.
void AsyncSignalSemaphoreWithIndex(int semaIndex) {
	if (write_port_etc(SqSem, semaIndex, 0, 0, B_TIMEOUT, 0) != B_NO_ERROR)
		fprintf(stderr, "\bUnable to signal squeak semaphore asynchronously.\n");
}
	
// this routine is executed during program exit
void Cleanup()
	{
	DEBUG_OUT("cleanup");

	ioShutdownAllModules();

	#if defined(PROFILE)
		bigtime_t total_t = 0;
	
		for (int i = 0; i < MAX_PROFILE_POINTS; i++)
			if (ProfElapsed[i] != -1)
				total_t += ProfElapsed[i];			
						
		if (total_t > 0)
			{
			printf("Profiling data\nid\tt/call\tMin. t\tMax. t\tCalls\t%% t\tName\n");		
			for (int i = 0; i < MAX_PROFILE_POINTS; i++)
				if (ProfCount[i] != -1)
					printf("%d\t%Ld\t%Ld\t%Ld\t%d\t%.1f\t%s\n", 
						i, ProfElapsed[i] / ProfCount[i], ProfMin[i], ProfMax[i], ProfCount[i], ProfElapsed[i] * 100.0 / total_t, ProfNames[i]);
			printf("Total time in profiled areas: %Ld\n", total_t);
			}
	#endif

	DEBUG_OUT("cleanup/ret");
	}

// parse command line arguments	
void ParseArguments(int argc, char **argv)
	{
	int cur_arg = 1;
	
	// parse argments
	for (; cur_arg < argc; cur_arg++)
		{		
		// flag
		if (argv[cur_arg][0] == '-')
			{
			if (strncmp(argv[cur_arg], "-m", 2) == 0)
				{
				// set amount of memory to allocate
				MemorySize = atoi(&argv[cur_arg][2]) << 20;
				if (MemorySize <= 0) MemorySize = DefaultMemorySize;
				}
			else if (strncmp(argv[cur_arg], "-win:d", 6) == 0)
				{
				if (ForceBWindow)
					printf("Force BWindow already specified, argument ignored.\n");
				// force direct window display
				ForceDirectWindow = true;
				}
			else if (strncmp(argv[cur_arg], "-win:b", 6) == 0)
				{
				if (ForceDirectWindow)
					printf("Force DirectWindow already specified, argument ignored.\n");
				// force direct window display
				ForceBWindow = true;
				}
			else
				{
				ShowAlert(argv[cur_arg]);
				ShowUsage();
				exit(-3);
				}
			}
		else
			{
			// image name
 			// check argument length to avoid segment violation on long argument
 			if (strlen(argv[cur_arg]) >= B_PATH_NAME_LENGTH)
 				fprintf(stderr, "Image name parameter longer than maximum BeOS path name length.\n");
 			else {
 				strcpy(ShortImageName, argv[cur_arg]);
				PubImageArg = cur_arg;
 			}
			break;
			}
		}
	
	}

// add a profiling timing entry	
void ProfileEntry(int id, bigtime_t time)
	{
	if (ProfElapsed[id] == -1)
		{		
		ProfElapsed[id] = time;
		ProfMin[id] = time;
		ProfMax[id] = time;
		ProfCount[id] = 1;
		}
	else
		{		
		ProfElapsed[id] += time;
		ProfMin[id] = time < ProfMin[id] ? time : ProfMin[id];
		ProfMax[id] = time > ProfMax[id] ? time : ProfMax[id];
		ProfCount[id]++;
		}		
	}
	
// show an error alert	
void ShowAlert(const char *err)
	{
	DEBUG_OUT("show_alert");
	BAlert *alert = new BAlert("Error", err,
		"Exit", NULL, NULL, B_WIDTH_FROM_LABEL, B_STOP_ALERT); 
   	alert->Go();
	DEBUG_OUT("show_alert/ret");
	} 
	
// show proper command line usage	
void ShowUsage()
	{
	printf("BeSqueakXXX [vmFlags] imageName [imageFlags]\n\n"
			"Where XXX is one of the following: x86, PPC\n\n"
			"vmFlags can be any of the following:\n"
			"\t-mX\twhere X is the number of megabytes of memory to \n\t\tallocate to hold the image (32 is the default)\n"
			"\t-win:X\twhere X is the either b (display using BWindow)\n\t\tor d (display using BDirectWindow).  d is the default\n");
	exit(-6);
	}
	
// this routine is called by the interpreter thread to signal the squeak
// semaphores pending from calls to AsyncSignalSemaphoreWithIndex
void SignalSqSemaphores()
	{
	int32 semaIndex;
	
	while (read_port_etc(SqSem, &semaIndex, 0, 0, B_TIMEOUT, 0) == 0)
		signalSemaphoreWithIndex((int)semaIndex);
	}

// this function is spawned as the squeak interpreter thread
int32 StartInterpret(void *)
	{
	DEBUG_OUT("StartInterpret");
	sqImageFile f;

	/* read the image file and allocate memory for Squeak heap */
	f = sqImageFileOpen(imageName, "rb");
	if (f == NULL) {
		const char *err = "Could not open the Squeak image file '%s'\n\n";
		char fmterr[512];
		sprintf(fmterr, err, imageName);
		/* give a error message if image file is not found */
		ShowAlert(fmterr);
		BMessage msg(B_QUIT_REQUESTED);
		be_app->PostMessage(&msg);
		DEBUG_OUT("StartInterpret/ret");
		exit_thread(B_OK);
	}
	readImageFromFileHeapSizeStartingAt(f, MemorySize, 0);  
	sqImageFileClose(f);
    
    // initialize security subsystem
    ioInitSecurity();
    
 	// decide whether to use BDirectWindow or BWindow based on user request and support for BDirectWindow
 	UsingDirectWindow = (BDirectWindow::SupportsWindowMode() || ForceDirectWindow) && (!ForceBWindow);

	BRect WindowRect = CalcWindowRect(fullScreenFlag);
	if (UsingDirectWindow) 
		// do the direct window thing
		SqWindow = new SqueakDirectWindow(WindowRect);
	else 
		// slow, but supported everywhere
		SqWindow = new SqueakWindow(WindowRect);

	SqWindow->SetTitle(ShortImageName);
	// show interpreter window
	SqWindow->Show();
	
	// start interpreter
	interpret();
	
	DEBUG_OUT("StartInterpret/ret");

	return 0;
	}
	
//////////
// main //
//////////
	
int main(int argc, char **argv) 
	{
	DEBUG_OUT("main");
	
	// put argc & argv in global variables -- there probably are already global vars defined by the compiler
 	// is this covered by ANSI C/C++ standards?
	Pubargc = &argc;
	Pubargv = &argv;

	ProgName = argv[0];
	imageName[0] = ShortImageName[0] = VMPath[0] = 0;
	
	// default image name	
	strcpy(ShortImageName, "squeak.image");
				
	// instatiate the app & run it	
	SqApp = new SqueakApp;
	
	SqApp->Run();		
	
	delete SqApp;
			
	DEBUG_OUT("main/ret");
	return 0;
}
