// this file contains: BeOS DisplayBitmap copying support.  These routines copy squeak images to either a BDirectWindow or a BBitmap.
//
// authors: Trey Bordeau, Colin Sarsfield
//
// Please read the file named License for license information.
// 

#include "sqBeCopyBits.h"
#include "sqBeGlobal.h"
#include <ByteOrder.h>
#include <stdio.h>
#include <stdlib.h>

// macros to convert from one graphics representation to another (i.e. 8-bit to 32-bit)
#define ALPHA_CHAN_15           0x8000 // alpha channel for 15-bit display modes
#define ALPHA_CHAN_32           0xff000000 // alpha channel for 32-bit display modes
//#define CONVERT_15_TO_16_S32(x)	((((x) & (0x7fe0 << 16)) >> 15) | (((x) & (0x1f << 16)) >> 16) | (((x) & 0x7fe0) << 17) | (((x) & 0x1f)) << 16)
#define CONVERT_15_TO_16(x)		((((x) & 0x7fe0) << 1) | ((x) & 0x1f)) // convert 15-bit to 16-bit (16-bit has an extra bit for green)
#define CONVERT_32_TO_15NA(x)	((((x) & 0xf8) >> 3) | (((x) & 0xf800) >> 6) | (((x) & 0xf80000) >> 9)) // convert 32-bit to 15-bit clearing the alpha channel
#define CONVERT_32_TO_15(x)		(ALPHA_CHAN_15 | CONVERT_32_TO_15NA(x)) // convert 32-bit to 15-bit setting the alpha channel
#define CONVERT_32_TO_16(x)		((((x) & 0xf8) >> 3) | (((x) & 0xfc00) >> 5) | (((x) & 0xf80000) >> 8)) // convert 32-bit to 16-bit
#define CONVERT_15_TO_32(x)		(ALPHA_CHAN_32 | (((x) & 0x7c00) << 9) | (((x) & 0x3e0) << 6) | (((x) & 0x1f) << 3)) // convert 15-bit to 32-bit setting the alpha channel
#define CONVERT_8_TO_32(x)		((cm->color_list[(x)].alpha << 24) | (cm->color_list[(x)].red << 16) | \
								(cm->color_list[(x)].green << 8) | (cm->color_list[(x)].blue)) // convert squeak 8-bit to 32-bit
#define CONVERT_8_TO_15(x)		(((cm->color_list[(x)].alpha & 0x80) << 8) | ((cm->color_list[(x)].red & 0xf8) << 7) | \
								((cm->color_list[(x)].green & 0xf8) << 2) | ((cm->color_list[(x)].blue & 0xf8) >> 3)) // convert squeak 8-bit to 15-bit
#define CONVERT_8_TO_16(x)		(((cm->color_list[(x)].red & 0xf8) << 8) | \
								((cm->color_list[(x)].green & 0xfc) << 3) | ((cm->color_list[(x)].blue & 0xf8) >> 3)) // convert squeak 8-bit to 16-bit

uint8 SqueakToBeColorMap8[256]; // array for mapping squeak 8-bit colors to BeOS 8-bit colors
uint16 SqueakToBeColorMap15[256]; // array for mapping squeak 8-bit colors to BeOS 15-bit colors
uint16 SqueakToBeColorMap16[256]; // array for mapping squeak 8-bit colors to BeOS 16-bit colors
uint32 SqueakToBeColorMap32[256]; // array for mapping squeak 8-bit colors to BeOS 32-bit colors

// this routine builds the contents of the SqueakToBeColorMapXX arrays based on the current screen's color map
void InitSqueakToBeColorMap(BScreen *screen) {
	DEBUG_OUT("InitSqueakToBeColorMap");
	const color_map *cm = screen->ColorMap();	

	// we use this macro to define mappings from 8-bit squeak color to 8-bit BeOS screen color
	#define SetColorEntry(i, r, g, b)	SqueakToBeColorMap8[i] = screen->IndexForColor(r >> 8, g >> 8, b >> 8)
	
	/* 1-bit colors (monochrome) */
	SetColorEntry(0, 65535, 65535, 65535);	/* white or transparent */
	SetColorEntry(1,     0,     0,     0);	/* black */

	/* additional colors for 2-bit color */
	SetColorEntry(2, 65535, 65535, 65535);	/* opaque white */
	SetColorEntry(3, 32768, 32768, 32768);	/* 1/2 gray */

	/* additional colors for 4-bit color */
	SetColorEntry( 4, 65535,     0,     0);	/* red */
	SetColorEntry( 5,     0, 65535,     0);	/* green */
	SetColorEntry( 6,     0,     0, 65535);	/* blue */
	SetColorEntry( 7,     0, 65535, 65535);	/* cyan */
	SetColorEntry( 8, 65535, 65535,     0);	/* yellow */
	SetColorEntry( 9, 65535,     0, 65535);	/* magenta */
	SetColorEntry(10,  8192,  8192,  8192);	/* 1/8 gray */
	SetColorEntry(11, 16384, 16384, 16384);	/* 2/8 gray */
	SetColorEntry(12, 24576, 24576, 24576);	/* 3/8 gray */
	SetColorEntry(13, 40959, 40959, 40959);	/* 5/8 gray */
	SetColorEntry(14, 49151, 49151, 49151);	/* 6/8 gray */
	SetColorEntry(15, 57343, 57343, 57343);	/* 7/8 gray */

	/* additional colors for 8-bit color */
	/* 24 more shades of gray (does not repeat 1/8th increments) */
	SetColorEntry(16,  2048,  2048,  2048);	/*  1/32 gray */
	SetColorEntry(17,  4096,  4096,  4096);	/*  2/32 gray */
	SetColorEntry(18,  6144,  6144,  6144);	/*  3/32 gray */
	SetColorEntry(19, 10240, 10240, 10240);	/*  5/32 gray */
	SetColorEntry(20, 12288, 12288, 12288);	/*  6/32 gray */
	SetColorEntry(21, 14336, 14336, 14336);	/*  7/32 gray */
	SetColorEntry(22, 18432, 18432, 18432);	/*  9/32 gray */
	SetColorEntry(23, 20480, 20480, 20480);	/* 10/32 gray */
	SetColorEntry(24, 22528, 22528, 22528);	/* 11/32 gray */
	SetColorEntry(25, 26624, 26624, 26624);	/* 13/32 gray */
	SetColorEntry(26, 28672, 28672, 28672);	/* 14/32 gray */
	SetColorEntry(27, 30720, 30720, 30720);	/* 15/32 gray */
	SetColorEntry(28, 34815, 34815, 34815);	/* 17/32 gray */
	SetColorEntry(29, 36863, 36863, 36863);	/* 18/32 gray */
	SetColorEntry(30, 38911, 38911, 38911);	/* 19/32 gray */
	SetColorEntry(31, 43007, 43007, 43007);	/* 21/32 gray */
	SetColorEntry(32, 45055, 45055, 45055);	/* 22/32 gray */
	SetColorEntry(33, 47103, 47103, 47103);	/* 23/32 gray */
	SetColorEntry(34, 51199, 51199, 51199);	/* 25/32 gray */
	SetColorEntry(35, 53247, 53247, 53247);	/* 26/32 gray */
	SetColorEntry(36, 55295, 55295, 55295);	/* 27/32 gray */
	SetColorEntry(37, 59391, 59391, 59391);	/* 29/32 gray */
	SetColorEntry(38, 61439, 61439, 61439);	/* 30/32 gray */
	SetColorEntry(39, 63487, 63487, 63487);	/* 31/32 gray */

	/* The remainder of color table defines a color cube with six steps
	for each primary color. Note that the corners of this cube repeat
	previous colors, but simplifies the mapping between RGB colors and
	color map indices. This color cube spans indices 40 through 255.
	*/

	for (int r= 0; r < 6; r++)
		for (int g= 0; g < 6; g++)
			for (int b= 0; b < 6; b++) {
				int i = 40 + ((36 * r) + (6 * b) + g);
				if (i > 255) fprintf(stderr, "index out of range in color table compuation");
				SetColorEntry(i, r * 65535 / 5, g * 65535 / 5, b * 65535 / 5);
			}

	// after we have the 8-bit squeak to BeOS mapping, the other maps can be looked
	// up using the BeOS 8-bit color map
	for (int j = 0; j < 256; j++) {
		SqueakToBeColorMap15[j] = CONVERT_8_TO_15(SqueakToBeColorMap8[j]);
		SqueakToBeColorMap16[j] = CONVERT_8_TO_16(SqueakToBeColorMap8[j]);
		SqueakToBeColorMap32[j] = CONVERT_8_TO_32(SqueakToBeColorMap8[j]);
	}

	DEBUG_OUT("InitSqueakToBeColorMap/ret");
	}

// pointer-to-member function
typedef void (CopyBits::*rblter)();

CopyBits::CopyBits(const color_map *aCM, int aSrcDepth, color_space aDstCS, 
		int aWidth, int aHeight, void *aSrc, void *aDst,
		int aDstXOff, int aDstYOff, int aBeBytesPerRow) {
	// we use a pointer to a member function so that we can avoid the overhead of a case statement or array access later
	// a minor saving, but an easy enough one to implement
	static rblter blter_lookup[2][4][6] = {
		{ {&CopyBits::Copy32to8, &CopyBits::Copy16to8, &CopyBits::Copy8to8, &CopyBits::Copy4to8, &CopyBits::Copy2to8, &CopyBits::Copy1to8},
		{&CopyBits::Copy32to15, &CopyBits::Copy16to15, &CopyBits::Copy8to15, &CopyBits::Copy4to15, &CopyBits::Copy2to15, &CopyBits::Copy1to15},
		{&CopyBits::Copy32to16, &CopyBits::Copy16to16, &CopyBits::Copy8to16, &CopyBits::Copy4to16, &CopyBits::Copy2to16, &CopyBits::Copy1to16},
		{&CopyBits::Copy32to32, &CopyBits::Copy16to32, &CopyBits::Copy8to32, &CopyBits::Copy4to32, &CopyBits::Copy2to32, &CopyBits::Copy1to32} },
		{ {&CopyBits::Copy32to8, &CopyBits::Swap16to8, &CopyBits::Swap8to8, &CopyBits::Swap4to8, &CopyBits::Swap2to8, &CopyBits::Swap1to8},
		{&CopyBits::Copy32to15, &CopyBits::Swap16to15, &CopyBits::Swap8to15, &CopyBits::Swap4to15, &CopyBits::Swap2to15, &CopyBits::Swap1to15},
		{&CopyBits::Copy32to16, &CopyBits::Swap16to16, &CopyBits::Swap8to16, &CopyBits::Swap4to16, &CopyBits::Swap2to16, &CopyBits::Swap1to16},
		{&CopyBits::Copy32to32, &CopyBits::Swap16to32, &CopyBits::Swap8to32, &CopyBits::Swap4to32, &CopyBits::Swap2to32, &CopyBits::Swap1to32} }
		};
	int fun_idx, endian_idx;
		
	// store parameters into members for future use
	SrcDepth = aSrcDepth;
	DstCS = aDstCS;
	Width = aWidth;
	Height = aHeight;
	DstXOff = aDstXOff;
	DstYOff = aDstYOff;
	cm = aCM;
	Src = static_cast<unsigned char *>(aSrc);
	Dst = static_cast<unsigned char *>(aDst);	
	
	//printf("DstXOff: %d, DstYOff: %d\n", DstXOff, DstYOff);
	
	//printf("SrcDepth: %d, DstCS: %d, Width: %d, Height: %d\n", SrcDepth, DstCS, Width, Height);

	// 1 << tx is the number of pixels per 32-bit word
	// tx = log2(32 / aSrcDepth)
	// if (aSrcDepth == 32) tx = 0;
	// if (aSrcDepth == 16) tx = 1;
	// if (aSrcDepth == 8) tx = 2;
	// if (aSrcDepth == 4) tx = 3;
	// if (aSrcDepth == 2) tx = 4;
	// if (aSrcDepth == 1) tx = 5;
	// basically tx is a number that is often handy for calculating the length of a bitmap row in 32-bit words
	tx = 6;
	while (aSrcDepth)  {
		tx--;
		aSrcDepth >>= 1;
	}

	// DstTx is for the BeOS side whereas tx is for the Squeak side, note that DstTx follows the same rules as tx
	switch (DstCS)
		{
		case B_CMAP8:
			DstDepth = 8;
			DstTx = 2;
			fun_idx = 0;
			break;
		case B_RGB15_LITTLE:
		case B_RGBA15_LITTLE:
		case B_RGB15_BIG:
		case B_RGBA15_BIG:
			DstDepth = 15;
			DstTx = 1;
			fun_idx = 1;
			break;
		case B_RGB16_LITTLE:
		case B_RGB16_BIG:
			DstDepth = 16;
			DstTx = 1;
			fun_idx = 2;
			break;
		case B_RGB32_LITTLE:
		case B_RGB32_BIG:
		case B_RGBA32_LITTLE:
		case B_RGBA32_BIG:
			DstDepth = 32;
			DstTx = 0;
			fun_idx = 3;
			break;
		default:
			printf("Destination color space not supported\n");
			exit(-12);
		}

	//printf("Src: %d, Dst: %d, tx: %d, DstTx: %d\n", Src, Dst, tx, DstTx);

	// calculate the number of bytes per row for the source (Squeak) bitmap	
	squeakBytesPerRow = ((Width + (1 << tx) - 1) >> tx) << 2;

	// if the constructor is passed 0 for aBeBytesPerRow then CopyBits will calculate the number of bytes
	// per row in the destination (BeOS) bitmap and will note that it is a BBitmap and not display adapter memory
	beBytesPerRow = aBeBytesPerRow;
	IsBBitmap = aBeBytesPerRow == 0;
	
	// does Squeak (big endian) data need to be byte-swapped while copying to BeOS (can be either endian) data?	
	// DstCS & 0x1000 is true when the display adapter is big endian
	// (B_HOST_IS_BENDIAN && IsBBitmap) is true when a BBitmap is being used on a big endian machine (PPC BBitmap)
	// swapping of data is only required for little endian machines (Intel BBitmap)
	// note that when BDirectWindow is used the display adapter card determines whether the destination
	// bitmap is big or little endian (both Intel & PPC)
	if (IsBBitmap)
		endian_idx = B_HOST_IS_BENDIAN ? 0 : 1;
	else
		endian_idx = (DstCS & 0x1000) ? 0 : 1;
	
	// point to appropriate bitmap copying method based on:
	// 1. endianness of destination (BeOS) bitmap
	// 2. pixel size of destination (BeOS) bitmap
	// 3. pixel size of source (Squeak) bitmap
	RectBlter = blter_lookup[endian_idx][fun_idx][tx];
	//printf("squeakBytesPerRow: %d, beBytesPerRow: %d\n", squeakBytesPerRow, beBytesPerRow);
	}		
	
int CopyBits::BeWidth()
	{
	// the maximum number of pixels wide the squeak bitmap could be given how many bytes per row of storage it has
	// in other words:
	// r = squeakBytesPerRow * 8 / bitsPerPixel
	int r = ((squeakBytesPerRow << 3) >> (5 - tx));
	//printf("%d/%d\n", squeakBytesPerRow, r);
	return r;
	}

// used to loop through pixels by column in 1-bit, 2-bit and 4-bit bitmaps
#define COL_LOOP_BIT(conv_bit)		{ sqc = (sqr << 3) + sqbegin_bit; bits = Dst + br + bbegin; sqbits = Src + (sqc >> 3); int bit_over = sqc & 0x7; \
								  for (bitsend = Dst + (br + bend); bits <= bitsend; sqc += sqinc_bit, bits += binc, sqbits = Src + (sqc >> 3), bit_over = sqc & 0x7) conv_bit }
// used to loop through pixel rows in all bitmaps
#define ROW_LOOP 				int br, sqr, sqc; \
								  unsigned char *bits, *sqbits, *bitsend; \
								  for (sqr = sqrbegin, br = brbegin; sqr <= sqrend; \
								  sqr += squeakBytesPerRow, br += beBytesPerRow)
// used to loop through pixels by column in 8-bit, 16-bit and 32-bit bitmaps
#define COL_LOOP(conv)			{ sqc = sqr + sqbegin; \
								  bits = Dst + br + bbegin; \
								  sqbits = Src + sqc; \
								  for (bitsend = Dst + (br + bend); \
								  bits <= bitsend; bits += binc, sqbits += sqinc) conv }

#define SWAP_NIBBLE(x)			(((x) >> (bit_over ^ 4)) & 0xf)  // extracts a nibble (4-bit) field from a byte reversing endianness
#define SWAP_2_BIT(x)			(((x) >> (bit_over ^ 6)) & 3)  // extracts 2-bit field from a byte reversing endianness
#define SWAP_1_BIT(x)			(((x) >> (bit_over ^ 7)) & 1)  // extracts 1-bit field from a byte reversing endianness

#define COPY_NIBBLE(x)			(((x) >> bit_over) & 0xf) // extracts a nibble (4-bit) field from a byte
#define COPY_2_BIT(x)			(((x) >> bit_over) & 3) // extracts 2-bit field from a byte
#define COPY_1_BIT(x)			(((x) >> bit_over) & 1) // extracts 1-bit field from a byte

// macros for the inside loop of all the big endian copying routines
#define COPY_1_TO_8				{ *bits = SqueakToBeColorMap8[COPY_1_BIT(*sqbits)]; }
#define COPY_1_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[COPY_1_BIT(*sqbits)]; }
#define COPY_1_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[COPY_1_BIT(*sqbits)]; }
#define COPY_1_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[COPY_1_BIT(*sqbits)]; }
#define COPY_2_TO_8				{ *bits = SqueakToBeColorMap8[COPY_2_BIT(*sqbits)]; }
#define COPY_2_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[COPY_2_BIT(*sqbits)]; }
#define COPY_2_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[COPY_2_BIT(*sqbits)]; }
#define COPY_2_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[COPY_2_BIT(*sqbits)]; }
#define COPY_4_TO_8				{ *bits = SqueakToBeColorMap8[COPY_NIBBLE(*sqbits)]; }
#define COPY_4_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[COPY_NIBBLE(*sqbits)]; }
#define COPY_4_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[COPY_NIBBLE(*sqbits)]; }
#define COPY_4_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[COPY_NIBBLE(*sqbits)]; }
#define COPY_8_TO_8				{ *bits = SqueakToBeColorMap8[*sqbits]; }
#define COPY_8_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[*sqbits]; }
#define COPY_8_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[*sqbits]; }
#define COPY_8_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[*sqbits]; }
#define COPY_16_TO_8			{ *bits = cm->index_map[*((uint16 *)sqbits) & 0x7fff]; }
#define COPY_16_TO_15			{ *((short *)bits) = ALPHA_CHAN_15 | *((short *)sqbits); }
#define COPY_16_TO_16			{ *((short *)bits) = CONVERT_15_TO_16(*((short *)sqbits)); }
#define COPY_16_TO_32			{ *((int *)bits) = CONVERT_15_TO_32((int)*((short *)sqbits)); }
#define COPY_32_TO_8			{ *bits = cm->index_map[CONVERT_32_TO_15NA(*((uint32 *)sqbits))]; }
#define COPY_32_TO_15			{ *((short *)bits) = CONVERT_32_TO_15(*((int *)sqbits)); }
#define COPY_32_TO_16			{ *((short *)bits) = CONVERT_32_TO_16(*((int *)sqbits)); }
#define COPY_32_TO_32			{ *((int *)bits) = (*((int *)sqbits)) | ALPHA_CHAN_32; }

// macros for the inside loop of all the little endian copying routines
#define SWAP_1_TO_8				{ *bits = SqueakToBeColorMap8[SWAP_1_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_1_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[SWAP_1_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_1_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[SWAP_1_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_1_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[SWAP_1_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_2_TO_8				{ *bits = SqueakToBeColorMap8[SWAP_2_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_2_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[SWAP_2_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_2_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[SWAP_2_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_2_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[SWAP_2_BIT(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_4_TO_8				{ *bits = SqueakToBeColorMap8[SWAP_NIBBLE(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_4_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[SWAP_NIBBLE(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_4_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[SWAP_NIBBLE(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_4_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[SWAP_NIBBLE(*((unsigned char *)((int)sqbits ^ 3)))]; }
#define SWAP_8_TO_8				{ *bits = SqueakToBeColorMap8[*((unsigned char *)((int)sqbits ^ 3))]; }
#define SWAP_8_TO_15			{ *((uint16 *)bits) = SqueakToBeColorMap15[*((unsigned char *)((int)sqbits ^ 3))]; }
#define SWAP_8_TO_16			{ *((uint16 *)bits) = SqueakToBeColorMap16[*((unsigned char *)((int)sqbits ^ 3))]; }
#define SWAP_8_TO_32			{ *((int *)bits) = SqueakToBeColorMap32[*((unsigned char *)((int)sqbits ^ 3))]; }
#define SWAP_16_TO_8			{ *bits = cm->index_map[*((uint16 *)((int)sqbits ^ 2)) & 0x7fff]; }
#define SWAP_16_TO_15			{ *((short *)bits) = ALPHA_CHAN_15 | *((short *)((int)sqbits ^ 2)); }
#define SWAP_16_TO_16			{ *((short *)bits) = CONVERT_15_TO_16(*((short *)((int)sqbits ^ 2))); }
#define SWAP_16_TO_32			{ *((int *)bits) = CONVERT_15_TO_32((int)*((short *)((int)sqbits ^ 2))); }

#define DUD						{ printf("This copybits operation is not yet supported.\n"); exit(-12); }

// the actual member functions that implement the bitmap copying for big endian destination data (BeOS)
void CopyBits::Copy1to8() { ROW_LOOP COL_LOOP_BIT(COPY_1_TO_8) }
void CopyBits::Copy1to15() { ROW_LOOP COL_LOOP_BIT(COPY_1_TO_15) }
void CopyBits::Copy1to16() { ROW_LOOP COL_LOOP_BIT(COPY_1_TO_16) }
void CopyBits::Copy1to32() { ROW_LOOP COL_LOOP_BIT(COPY_1_TO_32) }
void CopyBits::Copy2to8() { ROW_LOOP COL_LOOP_BIT(COPY_2_TO_8) }
void CopyBits::Copy2to15() { ROW_LOOP COL_LOOP_BIT(COPY_2_TO_15) }
void CopyBits::Copy2to16() { ROW_LOOP COL_LOOP_BIT(COPY_2_TO_16) }
void CopyBits::Copy2to32() { ROW_LOOP COL_LOOP_BIT(COPY_2_TO_32) }
void CopyBits::Copy4to8() { ROW_LOOP COL_LOOP_BIT(COPY_4_TO_8) }
void CopyBits::Copy4to15() { ROW_LOOP COL_LOOP_BIT(COPY_4_TO_15) }
void CopyBits::Copy4to16() { ROW_LOOP COL_LOOP_BIT(COPY_4_TO_16) }
void CopyBits::Copy4to32() { ROW_LOOP COL_LOOP_BIT(COPY_4_TO_32) }
void CopyBits::Copy8to8() { ROW_LOOP COL_LOOP(COPY_8_TO_8) }
void CopyBits::Copy8to15() { ROW_LOOP COL_LOOP(COPY_8_TO_15) }
void CopyBits::Copy8to16() { ROW_LOOP COL_LOOP(COPY_8_TO_16) }
void CopyBits::Copy8to32() { ROW_LOOP COL_LOOP(COPY_8_TO_32) }
void CopyBits::Copy16to8() { ROW_LOOP COL_LOOP(COPY_16_TO_8) }
void CopyBits::Copy16to15() { ROW_LOOP COL_LOOP(COPY_16_TO_15) }
void CopyBits::Copy16to16() { ROW_LOOP COL_LOOP(COPY_16_TO_16) }
void CopyBits::Copy16to32() { ROW_LOOP COL_LOOP(COPY_16_TO_32) }
void CopyBits::Copy32to8() { ROW_LOOP COL_LOOP(COPY_32_TO_8) }
void CopyBits::Copy32to15() { ROW_LOOP COL_LOOP(COPY_32_TO_15) }
void CopyBits::Copy32to16() { ROW_LOOP COL_LOOP(COPY_32_TO_16) }
void CopyBits::Copy32to32() { ROW_LOOP COL_LOOP(COPY_32_TO_32) }

// the actual member functions that implement the bitmap copying for little endian destination data (BeOS)
void CopyBits::Swap1to8() { ROW_LOOP COL_LOOP_BIT(SWAP_1_TO_8) }
void CopyBits::Swap1to15() { ROW_LOOP COL_LOOP_BIT(SWAP_1_TO_15) }
void CopyBits::Swap1to16() { ROW_LOOP COL_LOOP_BIT(SWAP_1_TO_16) }
void CopyBits::Swap1to32() { ROW_LOOP COL_LOOP_BIT(SWAP_1_TO_32) }
void CopyBits::Swap2to8() { ROW_LOOP COL_LOOP_BIT(SWAP_2_TO_8) }
void CopyBits::Swap2to15() { ROW_LOOP COL_LOOP_BIT(SWAP_2_TO_15) }
void CopyBits::Swap2to16() { ROW_LOOP COL_LOOP_BIT(SWAP_2_TO_16) }
void CopyBits::Swap2to32() { ROW_LOOP COL_LOOP_BIT(SWAP_2_TO_32) }
void CopyBits::Swap4to8() { ROW_LOOP COL_LOOP_BIT(SWAP_4_TO_8) }
void CopyBits::Swap4to15() { ROW_LOOP COL_LOOP_BIT(SWAP_4_TO_15) }
void CopyBits::Swap4to16() { ROW_LOOP COL_LOOP_BIT(SWAP_4_TO_16) }
void CopyBits::Swap4to32() { ROW_LOOP COL_LOOP_BIT(SWAP_4_TO_32) }
void CopyBits::Swap8to8() { ROW_LOOP COL_LOOP(SWAP_8_TO_8) }
void CopyBits::Swap8to15() { ROW_LOOP COL_LOOP(SWAP_8_TO_15) }
void CopyBits::Swap8to16() { ROW_LOOP COL_LOOP(SWAP_8_TO_16) }
void CopyBits::Swap8to32() { ROW_LOOP COL_LOOP(SWAP_8_TO_32) }
void CopyBits::Swap16to8() { ROW_LOOP COL_LOOP(SWAP_16_TO_8) }
void CopyBits::Swap16to15() { ROW_LOOP COL_LOOP(SWAP_16_TO_15) }
void CopyBits::Swap16to16() { ROW_LOOP COL_LOOP(SWAP_16_TO_16) }
void CopyBits::Swap16to32() { ROW_LOOP COL_LOOP(SWAP_16_TO_32) }

// this little function copies a given rectangle from the source bitmap (Squeak) to the destination bitmap (BeOS)				
void CopyBits::CopyRect(int left, int top, int right, int bottom) {
	#if defined(PROFILE)
		bigtime_t now = real_time_clock_usecs();
	#endif
	//printf("(%d, %d)-(%d, %d)\n", left, top, right, bottom);
	//printf("left: %d, beBytesPerRow: %d, DstXOff: %d\n", left, beBytesPerRow, DstXOff);
	//printf("right: %d, tx: %d, DstTx: %d, top: %d, DstYOff: %d\n", right, tx, DstTx, top, DstYOff);
	//printf("squeakBytesPerRow: %d, Width: %d, Src: %d\n", squeakBytesPerRow, Width, (int)Src);
	
	// number of bytes per pixel (Squeak)
	sqinc = 4 >> tx;
	// number of bits per pixel (Squeak)
	sqinc_bit = 32 >> tx;
	// number of bytes per pixel (BeOS)
	binc = 4 >> DstTx;	 
	// the first bit in a row to copy (Squeak)
	sqbegin_bit = left << 5 >> tx;
	// the first byte in a row to copy (Squeak)
	sqbegin = (left << 2) >> tx;
	// note that sqbegin_bit & sqinc_bit are only used for 1-bit, 2-bit, and 4-bit squeak bitmaps

	// the first row to copy (Squeak)		
	sqrbegin = squeakBytesPerRow * top;
	// the last row to copy (Squeak)		
	sqrend = squeakBytesPerRow * bottom;
	// the first row to copy to (BeOS)		
	brbegin = beBytesPerRow * (top + DstYOff);
	// the first byte in a row to copy to (BeOS)
	bbegin = ((left + DstXOff) << 2) >> DstTx;
	// the last byte in a row to copy to (BeOS)
	bend = ((right + DstXOff) << 2) >> DstTx;
	// store the rectangles edges
	this->left = left;
	this->right = right;
	this->top = top;
	this->bottom = bottom;

	// call the appropriate copying member function - see constructor for details	
	(this->*RectBlter)();
		 
	PROF_ENT(1, real_time_clock_usecs() - now);
}	
	
void CopyBits::SetDst(void *aDst) {
	// set the destination (BeOS) bitmap pointer
	Dst = static_cast<unsigned char *>(aDst);	
}
	
void CopyBits::SetBeBytesPerRow(int aBeBytesPerRow) {
	// set the number of bytes per row for the destination (BeOS) bitmap
	beBytesPerRow = aBeBytesPerRow;
}
