// this file contains: BDirectWindow support
//
// authors: Trey Bordeau, Colin Sarsfield
//
// Please read the file named License for license information.
// 

#include "sqBeApp.h"
#include "sqBeDirectWindow.h"
#include "sqBeView.h"
#include "sqBeCopyBits.h"
#include "sqBeSharedWindow.h"
#include "sqBeGlobal.h"

#include <Locker.h>

// SqueakDirectWindow //

SqueakDirectWindow::SqueakDirectWindow(BRect frame)
		: BDirectWindow(frame, "BeSqueak", B_TITLED_WINDOW, B_OUTLINE_RESIZE) {
	DEBUG_OUT("SqueakDirectWindow::SqueakDirectWindow");
	fVisibleRegionChanged = false;
	fConnected = false;
	fConnectionDisabled = false;
	fLocker = new BLocker();
		
	// squeak display stuff
	fSqWidth = fSqHeight = fSqDepth = 0;
	fSqDirtyRegion.MakeEmpty();
	
	// add the views
	BRect bounds = Bounds();
	// see sqBeSharedWindow.cpp for details
	CreateViews(this, bounds);

	DEBUG_OUT("SqueakDirectWindow::SqueakDirectWindow/ret");
}

SqueakDirectWindow::~SqueakDirectWindow() {
	DEBUG_OUT("SqueakDirectWindow::~SqueakDirectWindow");

	fConnectionDisabled = true;      // Connection is dying
	Hide();
	Sync(); // not sure if this is still needed, but it works so I'll leave it alone for now
	SqWindow = 0;
	while (InDirectWindowDraw) snooze(1); // wait until direct drawing has finished before destructing the window
	delete fLocker;
	DEBUG_OUT("SqueakDirectWindow::~SqueakDirectWindow/ret");
}

// invalidate a region of the window
void SqueakDirectWindow::Invalidate(int left, int top, int right, int bottom) {
	DEBUG_OUT("SqueakDirectWindow::Invalidate");	
	if (DeferedRegLock.Lock()) {
		// rectangle is in Squeak coordinates
		clipping_rect r;
		r.left = left; r.top = top; r.right = right; r.bottom = bottom;
		DeferedRegion.Include(r);
		DeferedRegLock.Unlock();
	}
	DEBUG_OUT("SqueakDirectWindow::Invalidate/ret");
}

// direct window "message" handler
void SqueakDirectWindow::DirectConnected(direct_buffer_info *info) {
	DEBUG_OUT("DirectConnected");
	if (!fConnected && fConnectionDisabled)
		return;

	fLocker->Lock(); // this cannot happen at the same time as the direct draw loop

	switch(info->buffer_state & B_DIRECT_MODE_MASK) {
		case B_DIRECT_START:
			DEBUG_OUT("B_DIRECT_START");
			fConnected = true;
			// no "break" is intentional, the B_DIRECT_MODIFY code also applies to start
		case B_DIRECT_MODIFY: {
				DEBUG_OUT("B_DIRECT_MODIFY");
				// get window left, top (adjusted for BMenuBar) this will be used by direct draw code
				fBounds = info->window_bounds;
				fBounds.top += UsedTopSpace; 
				BPoint WindowOffset(fBounds.left, fBounds.top);
				// save important info about the display format/size
				fBits = (uint8 *) info->bits;
				fRowBytes = info->bytes_per_row;
				fFormat = info->pixel_format;
				// note that display region has changed since last direct draw
				fVisibleRegionChanged = true;
				// save the color map for translating Squeak colors to Be colors
				BScreen screen(this);	
				fColorMap = screen.ColorMap();
				// get the clipping region (which always has 0,0 as its left,top) and clip it to the squeak window minus the menu bar
				BRegion LastClipRegion = fClipRegion;
				fClipRegion.MakeEmpty();
				GetClippingRegion(&fClipRegion, &WindowOffset);
				BRegion sq_view_region;
				clipping_rect r;
				r.left = 0;
				r.top = 0;
				r.right = fBounds.right - fBounds.left;
				r.bottom = fBounds.bottom - fBounds.top;
				sq_view_region.Set(r);
				fClipRegion.IntersectWith(&sq_view_region);
				// fDiffRegion will show where to draw next time
				if (info->buffer_state & B_DIRECT_MODE_MASK == B_DIRECT_START) {
					BRegion fAddToDiffRegion;
					fAddToDiffRegion = fClipRegion;
					fAddToDiffRegion.Exclude(&LastClipRegion);			
					fDiffRegion.Include(&fAddToDiffRegion);
				} else
					fDiffRegion = fClipRegion;
			}
			break;
		case B_DIRECT_STOP:
			DEBUG_OUT("B_DIRECT_STOP");
			fConnected = false;
			break;
	}
	
	fLocker->Unlock();
	DEBUG_OUT("DirectConnected/ret");
}

void SqueakDirectWindow::FrameResized(float width, float height) {
	DEBUG_OUT("SqueakDirectWindow::FrameResized");
	// sqBeSharedWindow.cpp for details
	HandleFrameResized(this, width, height);
	BDirectWindow::FrameResized(width, height);
	DEBUG_OUT("SqueakDirectWindow::FrameResized/ret");
}

bool SqueakDirectWindow::QuitRequested() {
	// sqBeSharedWindow.cpp for details
	HandleQuitRequested();
	return(Quitting);
}	

void SqueakDirectWindow::WindowActivated(bool active) {
	DEBUG_OUT("WindowActivated");
	// make sure that the squeak mouse handler becomes inactive when the window's not active
	uint32 buttons;	
	WindowActive = active;
	if (active)
		SqView->MakeFocus();
	else
		SqView->GetMouse(&SavedMousePosition, &buttons);		

	BDirectWindow::WindowActivated(active);

	DEBUG_OUT("WindowActivated/ret");
}

void SqueakDirectWindow::DispatchMessage(BMessage *msg, BHandler *target) {
	DEBUG_OUT("DispatchMessage");
	// sqBeSharedWindow.cpp for details
	if (HandleDispatchMessage(this, msg, target))
		BDirectWindow::DispatchMessage(msg, target);
	DEBUG_OUT("DispatchMessage/ret");
}
	
void SqueakDirectWindow::UpdateSqueakDisplayInfo(int bits, int width, int height, int depth) {
	DEBUG_OUT("UpdateSqueakDisplayInfo");
	// used by the draw thread routine for setting squeak info dependent on the current squeak display mode
	// note that "tx" is well described in the sqCopyBit.cpp comments	
	if ((uint8 *)bits != fSqBits) 
		fSqBits = (uint8 *)bits;
		
	if ((fSqWidth != width) || (fSqDepth != depth)) {
		fSqWidth = width;
		fSqDepth = depth;
		int tx = 6;
		while (depth)  {
			tx--;
			depth >>= 1;
		}
		int t1 = 1 << tx;
		fSqRowBytes = ((width + t1 - 1) >> tx) << 2;
	}
	if (fSqHeight != height)
		fSqHeight = height;
	DEBUG_OUT("UpdateSqueakDisplayInfo/ret");
}

void SqueakDirectWindow::DrawThread() {
	DEBUG_OUT("DrawThread");
	BRegion leftovers;
	// when window hasn't been created yet don't draw yet
	if (!SqWindow) return;
	GoAheadAndDraw = false;
	InDirectWindowDraw = true;
	if (!fConnectionDisabled) {
		// lock window against DirectConnected
		fLocker->Lock();

		if (fConnected) {
			// serialize updates to the squeak dirty region
			DeferedRegLock.Lock();
			// if the window changed shape, use it as the update region
			if (fVisibleRegionChanged) {
				fSqDirtyRegion = fDiffRegion;
				fOldClipRegion = fClipRegion;
				fVisibleRegionChanged = false;
			}
			// otherwise, use the intersection of the window shape and the squeak dirty region
			else 
				fSqDirtyRegion = DeferedRegion;

			fSqDirtyRegion.IntersectWith(&fClipRegion);
				
			fDiffRegion.MakeEmpty();
			
			// clear out for next pass
			DeferedRegion.MakeEmpty();
			
			DeferedRegLock.Unlock();
			// NOTE: update region is in Squeak window coordinates

			// if there is something to update
			if (fSqDirtyRegion.CountRects() > 0) {
				int depth, width, height, dispBitsIndex;

				if (GetSqueakDisplayInfo(depth, width, height, dispBitsIndex)) {
					fprintf(stderr, "Error in DrawThread()\n");
					exit(-13);
				}	
					
				// setup some squeak display info variables before copy
				UpdateSqueakDisplayInfo(dispBitsIndex, width, height, depth);
				
				// setup copy bits object 
				CopyBits cp(fColorMap, fSqDepth, fFormat, fSqWidth, fSqHeight, 
					fSqBits, fBits, fBounds.left, fBounds.top, fRowBytes);

				int i, count;
				clipping_rect r;
				// clip to squeak display
				BRegion clipreg;
				leftovers = fSqDirtyRegion;
				clipping_rect SqueakDisplayArea;
				SqueakDisplayArea.left = 0;
				SqueakDisplayArea.top = 0;
				SqueakDisplayArea.right =  width - 1;
				SqueakDisplayArea.bottom = height - 1;
				leftovers.Exclude(SqueakDisplayArea);
				clipreg.Set(SqueakDisplayArea);
				fSqDirtyRegion.IntersectWith(&clipreg); 
				count = fSqDirtyRegion.CountRects();

				// copy each rectangle of data
				for (i = 0; i < count; i++) {
					// rectangle parameters are in squeak coordinates
					r = fSqDirtyRegion.RectAtInt(i);

					cp.CopyRect(r.left, r.top, r.right, r.bottom);
				}
			}
		}
		// done with our drawing
		fLocker->Unlock();
	}
	else
		printf("Connection disabled.\n");

	InDirectWindowDraw = false;
	DEBUG_OUT("DrawThread/ret");
	return;
}
