// this file contains: interpreter i/o support functions
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 

#include "sqBeApp.h"
#include "sqBeIOPrims.h"
#include "sqBeView.h"
#include "sqBeGlobal.h"
#include "sqBeCopyBits.h"
#include "sqBeSharedWindow.h"
#include "sqBeDirectWindow.h"

#include <Screen.h>
#include <Beep.h>
#include <Bitmap.h>
#include <Locker.h>

struct SqCursor SqueakCursor;
bool fDeferredUpdate = true;

void CopySqueakBits();
void DrawNow();
void InvalidateDeferedRegion();
void RetrieveKey();

// interpreter defined functions //

// disable the power manager if non-zero.  what power manager? :)
int ioDisablePowerManager(int disableIfNonZero) {
	return 0;
}

// set an asynchronous input semaphore index for events
int ioSetInputSemaphore(int semaIndex) {
	SqEventSem = semaIndex;
	
	if ((SqEvents = create_port(256, "squeak_event_port")) < B_NO_ERROR) {
		ShowAlert("Unable to create squeak event port.");	
		exit(-3);
	}

	return 1;
}

// retrieve the next input event from the OS
int ioGetNextEvent(sqInputEvent *evt) {
	if (SqEvents == 0) return 1;
	int32 msg_code;
	if (port_count(SqEvents) > 0)
		read_port_etc(SqEvents, &msg_code, (void*)evt, sizeof(sqInputEvent), /*B_TIMEOUT*/0, 0);
	return 1;
}

// beep
int ioBeep(void) {
	success(beep() == B_OK);
	return 0;
}

// exit the interpreter
// the death of the interpreter thread seems to get blocked frequently -- it gets killed 
// but its death could be cleaned up
int ioExit(void) {
	Quitting = true;

	BMessage msg(B_QUIT_REQUESTED);
	SqWindow->PostMessage(&msg);

	exit_thread(B_OK);

	return 0;
}

// copy squeak display to be display "NOW"
// if using a BWindow the display will be updated lazily as the BWindow drawing uses an
// intermediate BBitmap
// also if the defer display updates preference is checked the squeak data will be copied later
int ioForceDisplayUpdate(void)  {
	DEBUG_OUT("ioForceDisplayUpdate");
	if (Quitting) exit_thread(B_OK);

	if (fDeferredUpdate && !deferDisplayUpdates) {
		GoAheadAndDraw = true;
		DrawNow();
	}

	DEBUG_OUT("ioForceDisplayUpdate/ret");
	return 0;
}

// experimental printing support -- probably no longer used
// future versions of Squeak might support PostScript?
int ioFormPrint(int bitsAddr, int width, int height, int depth, double hScale, double vScale, int landscapeFlag) {
	/* experimental: print a form with the given bitmap, width, height, and depth at
	   the given horizontal and vertical scales in the given orientation */
	DEBUG_OUT("ioFormPrint");
	success(false);  /* stubbed out */
	DEBUG_OUT("ioFormPrint/ret");
	return 0;
}

// get squeak mouse buttons
int ioGetButtonState(void) {
	//DEBUG_OUT("ioGetButtonState");
	if (Quitting) exit_thread(B_OK);
	
	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();
	
	if (WindowActive) {
		//DEBUG_OUT("ioGetButtonState/ret");
		return ButtonState;
	} else {
		//DEBUG_OUT("ioGetButtonState/ret");
		return 0;
	}
}

// get next keystroke
int ioGetKeystroke(void) {
	DEBUG_OUT("ioGetKeystroke");

	if (Quitting) exit_thread(B_OK);

	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();

	RetrieveKey();
	if (HasPeekKey) {
		HasPeekKey = false;
		ButtonState = ((PeekKey >> 5) & 0xF8) | (ButtonState & 0x7);
		DEBUG_OUT("ioGetKeystroke/ret");
		return PeekKey;
	} else {
		DEBUG_OUT("ioGetKeystroke/ret");
		return -1;
	}
}

// reports which display depths are supported by the Squeak interpreter
int ioHasDisplayDepth(int bitsPerPixel) {
	switch (bitsPerPixel) {
		case 1: case 2: case 4: case 8: case 16: case 32:
			return 1;
		default: 
			return 0;
	}
}

// high resolution timer
int ioMicroMSecs(void) {
	if (Quitting) exit_thread(B_OK);
	return ((int)(real_time_clock_usecs() / 1000));
}

// return "current" mouse location
// when window is inactive or mouse is outside window, mouse location is constant
int ioMousePoint(void) {
	//DEBUG_OUT("ioMousePoint");
	if (Quitting) exit_thread(B_OK);

	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();
	  
	//DEBUG_OUT("ioMousePoint/ret");
	return (static_cast<int>(SavedMousePosition.x) << 16) 
			| (static_cast<int>(SavedMousePosition.y) & 0xFFFF);  
			// x is high 16 bits; y is low 16 bits
}

// medium res timer
int ioMSecs(void) {
	if (!Quitting) {
		// perform lazy squeak activities now
		SignalSqSemaphores();
		DrawNow();
	} else
		exit_thread(B_OK);

	return ((int)(real_time_clock_usecs() / 1000));
}

// peek at next keystroke
int ioPeekKeystroke(void) {
	DEBUG_OUT("ioPeekKeystroke");
	
	if (Quitting) exit_thread(B_OK);

	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();
	
	RetrieveKey();
	if (HasPeekKey) {
		ButtonState = ((PeekKey >> 5) & 0xF8) | (ButtonState & 0x7);
		DEBUG_OUT("ioPeekKeystroke/ret");
		return PeekKey;
	} else {
		DEBUG_OUT("ioPeekKeystroke/ret");
		return -1;
	}
}

// offical request from Squeak to allow program to do non Squeak stuff
// necessary on Mac for processing of window loop
int ioProcessEvents(void) {
	DEBUG_OUT("ioProcessEvents");
	if (Quitting) exit_thread(B_OK);

	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();
		
	DEBUG_OUT("ioProcessEvents/ret");
	return 1;
}

// request from Squeak to relinquish the CPU for "microSeconds".
// this is needed because the Squeak architecture is still polling
// hence it uses up tons of CPU time when doing nothing if this
// primitive is not implemented
int ioRelinquishProcessorForMicroseconds(int microSeconds) {
	DEBUG_OUT("ioRelinquishProcessorForMicroseconds");

	if (Quitting) exit_thread(B_OK);

	static int Latency = 1,
		       Calls = 1;

	bigtime_t start = real_time_clock_usecs();
	
	// perform lazy squeak activities now
	SignalSqSemaphores();
	DrawNow();

	// all this code just tries to get snooze to approximate it's request better
	// this is way too much work trying to second guess snooze, but for some
	// foolish reason Squeak animation seems to suffer if this routine is not
	// resonably precise
	int aveLatency = Latency / Calls;
	int snoozeReq = microSeconds - max_c(0, aveLatency);
	
	start = real_time_clock_usecs();
	snooze(snoozeReq > 0?snoozeReq:1);
	Calls++;
	Latency += real_time_clock_usecs() - start - snoozeReq;
	if (Calls >= (1 << 8)) {  // arbitrary selection
		Calls >>= 1;
		Latency >>= 1;
	}
	snooze(microSeconds);
	
	DEBUG_OUT("ioRelinquishProcessorForMicroseconds/ret");
	
	return microSeconds;
}

// return the screen depth of the OS display in bits per pixel
int ioScreenDepth(void) {
	BScreen screen;
	
	// I've erred on the side of over-completeness hear because it was easy...
	// most of these will probably never be returned by screen.ColorSpace()
	switch (screen.ColorSpace()) {
		case B_RGB32:
		case B_RGBA32:
			return 32;
		case B_RGB24:
			return 24;
		case B_RGB16:
		case B_RGB15:
		case B_RGBA15:
			return 16;
		case B_CMAP8:
		case B_GRAY8:
			return 8;
		case B_GRAY1:
			return 1;
		default: 
			return 8;
	}
}

// return size of Squeak display area
int ioScreenSize(void) {
	DEBUG_OUT("ioScreenSize");
	if (Quitting) exit_thread(B_OK);

	if (SqWindow->Lock()) {
		BRect rect = SqView->Bounds();
		SqWindow->Unlock();
		DEBUG_OUT("ioScreenSize/ret");
		return (static_cast<int>(rect.Width() + 1) << 16) |
			(static_cast<int>(rect.Height() + 1) & 0xFFFF);  /* w is high 16 bits; h is low 16 bits */
	} else {
		DEBUG_OUT("ioScreenSize/ret");
		return 0;
	}
}

// return seconds since the beginning of squeak time:
// jan 1, 1901
int ioSeconds(void) {
	DEBUG_OUT("ioSeconds");

	if (Quitting) exit_thread(B_OK);

	struct tm timeRec;
	time_t time1904, timeNow;

	/* start of ANSI epoch is midnight of Jan 1, 1904 */
	timeRec.tm_sec   = 0;
	timeRec.tm_min   = 0;
	timeRec.tm_hour  = 0;
	timeRec.tm_mday  = 1;
	timeRec.tm_mon   = 0;
	timeRec.tm_year  = 4;
	timeRec.tm_wday  = 0;
	timeRec.tm_yday  = 0;
	timeRec.tm_isdst = 0;
	time1904 = mktime(&timeRec);

	timeNow = time(NULL);

	/* Squeak epoch is Jan 1, 1901, 3 non-leap years earlier than ANSI one */
	DEBUG_OUT("ioSeconds/ret");

	return (timeNow - time1904) + (3 * 365 * 24 * 60 * 60);
}

// set cursor bitmap
int ioSetCursor(int cursorBitsIndex, int offsetX, int offsetY) {
	/* Old version; forward to new version. */
	DEBUG_OUT("ioSetCursor");
	DEBUG_OUT("ioSetCursor/ret");
	return ioSetCursorWithMask(cursorBitsIndex, 0, offsetX, offsetY);	
}

// set cursor bitmap with transparency mask
int ioSetCursorWithMask(int cursorBitsIndex, int cursorMaskIndex, int offsetX, int offsetY) {
	/* Set the 16x16 cursor bitmap. If cursorMaskIndex is nil, then make the mask the same as
	   the cursor bitmap. If not, then mask and cursor bits combined determine how cursor is
	   displayed:
			mask	cursor	effect
			 0		  0		transparent (underlying pixel shows through)
			 1		  1		opaque black
			 1		  0		opaque white
			 0		  1		invert the underlying pixel
	*/
	if (Quitting) exit_thread(B_OK);
	SqueakCursor.size = 16;
	SqueakCursor.depth = 1;
	SqueakCursor.x = -offsetX;
	SqueakCursor.y = -offsetY; 
	
	int i;
	
	#define SHIFT_CUR(x) (x >> 16)

	if (cursorMaskIndex == 0) {
		for (i = 0; i < 16; i++) {
			SqueakCursor.data[i] = SHIFT_CUR(checkedLongAt(cursorBitsIndex + (i << 2)));
			SqueakCursor.mask[i] = SHIFT_CUR(checkedLongAt(cursorBitsIndex + (i << 2)));
		}
	} else {
		for (i = 0; i < 16; i++) {
			SqueakCursor.data[i] = SHIFT_CUR(checkedLongAt(cursorBitsIndex + (i << 2)));
			SqueakCursor.mask[i] = SHIFT_CUR(checkedLongAt(cursorMaskIndex + (i << 2)));
		}
	}

	// endianness handling here	
	swap_data(B_INT16_TYPE, SqueakCursor.data, 32, B_SWAP_BENDIAN_TO_HOST);
	swap_data(B_INT16_TYPE, SqueakCursor.mask, 32, B_SWAP_BENDIAN_TO_HOST);
		
	if (SqWindow->Lock()) {
		if (CursorIsInView)
			SqApp->SetCursor(&SqueakCursor);
		SqWindow->Unlock();
	}
	
	return 1;
}

// allow squeak to change the current display mode -- not implemented
int ioSetDisplayMode(int width, int height, int depth, int fullScreenMode) {
	if (Quitting) exit_thread(B_OK);
	return 0;
}

// set to "full screen" mode
int ioSetFullScreen(int fullScreen) {
	//static int wasFullScreen = 0;
	DEBUG_OUT("ioSetFullScreen");

	if (Quitting) exit_thread(B_OK);
	
	BRect winRect = CalcWindowRect(fullScreen);

	SqWindow->MoveTo(winRect.left, winRect.top); 
	SqWindow->ResizeTo(winRect.Width(), winRect.Height());
		
	DEBUG_OUT("ioSetFullScreen/ret");
	return 1;
}

// copy squeak display bits to actual display
// this may occur lazily
int ioShowDisplay(int dispBitsIndex, int width, int height, int depth,
		int affectedL, int affectedR, int affectedT, int affectedB) {

	if (Quitting) exit_thread(B_OK);
	/* Don't draw empty/reverse areas */
	if (affectedR >= width) affectedR = width - 1;
	if (affectedB >= height) affectedB = height - 1;
	if (affectedR <= affectedL || affectedT >= affectedB) return 1;

	DEBUG_OUT("ioShowDisplay");

	#if defined(PROFILE)
		bigtime_t now = real_time_clock_usecs();
	#endif

	if (DeferedRegLock.Lock()) {
		clipping_rect drawRect;
		drawRect.left = affectedL;
		drawRect.top = affectedT;
		drawRect.right = affectedR;
		drawRect.bottom = affectedB;
		DeferedRegion.Include(drawRect);
		DeferedRegLock.Unlock();
	}
	
	PROF_ENT(6, real_time_clock_usecs() - now);
		
	GoAheadAndDraw = true;
	if (fDeferredUpdate && !deferDisplayUpdates)
		PROF_ENT(7, 0);
	else {
		DrawNow();
		PROF_ENT(9, 0);
	}

	return 0;
}

int plugInAllowAccessToFilePath(char *pathString, int pathStringLength) {
	return 1;
}

// support functions

BRect CalcWindowRect(int fullScreen) {
	BScreen screen; 
	BRect winRect;
	
	int width = (int)screen.Frame().Width(), 
		height = (int)screen.Frame().Height();
		
	int savedWidth = savedWindowSize >> 16,
		savedHeight = savedWindowSize & 0xffff;
		
	if (fullScreen)
		winRect = BRect(0, 0, width, height);
	else {
		if (savedHeight == 0)
			savedHeight = height - (height >> 3);
		else
			savedHeight = min_c(savedHeight, height - 4);

		if (savedWidth == 0)
			savedWidth = width - (width >> 3);
		else
			savedWidth = min_c(savedWidth, width - 4);
			
		savedHeight = max_c(savedHeight, 64);
		savedWidth = max_c(savedWidth, 64);

		int	left = (width - savedWidth) >> 1,
			top = (height - savedHeight) >> 1;
			
		winRect = BRect(left, top, left + savedWidth - 1, top + savedHeight - 1);

		savedWindowSize = (savedWidth << 16) | savedHeight;
	}
	
	fullScreenFlag = fullScreen;

	return winRect;
}

void CopySqueakBits() {
	color_space cs;
	static int oldWidth = -1, oldHeight = -1;
	static color_space oldCs = B_NO_COLOR_SPACE;
		
	int depth, width, height, dispBitsIndex;
	if (GetSqueakDisplayInfo(depth, width, height, dispBitsIndex)) {
		fprintf(stderr, "Error in CopySqueakBits()\n");
		exit(-13);
	}	
	
	GoAheadAndDraw = false;
	
	switch (depth) {
		case 1:
		case 2:
		case 4:
		case 8:
			cs = B_CMAP8;
			break;
		case 16:
			cs = B_RGB15;
			break;
		case 32:
			cs = B_RGB32;
			break;
		default:
			return;
	}
	
	BScreen screen(SqWindow);
	CopyBits cp(screen.ColorMap(), depth, cs, width, height, reinterpret_cast<void *>(dispBitsIndex), 0);	
	
	if (Scr == 0 || oldCs != cs || oldWidth != width || oldHeight != height) {
		if (Scr != 0) {
			delete Scr;	
			Scr = 0;
		}			

		Scr = new BBitmap(BRect(0, 0, cp.BeWidth() - 1, height - 1), cs); 
	
		oldCs = cs;
		oldWidth = width;
		oldHeight = height;
	}
		
	cp.SetBeBytesPerRow(Scr->BytesPerRow());
	cp.SetDst(Scr->Bits());
	int i, count;
	
	//DeferedRegion.PrintToStream();
	if (DeferedRegLock.Lock()) {
		for (i = 0, count = DeferedRegion.CountRects(); i < count; i++) {
			BRect rect;
			int32	top, left;
			int32	bottom, right;		

			// rectangle parameters are in squeak coordinates
			rect = DeferedRegion.RectAt(i);

			top = (int32)rect.top;
			left = (int32)rect.left;
			bottom = (int32)rect.bottom;
			right = (int32)rect.right;
		
			cp.CopyRect(left, top, right, bottom);
		}
		DeferedRegLock.Unlock();
	}

	return;
}

void DrawNow() {
	if (GoAheadAndDraw) {
		#if defined(PROFILE)
			bigtime_t now = real_time_clock_usecs();
		#endif

		if (UsingDirectWindow)
			((SqueakDirectWindow *)SqWindow)->DrawThread();
		else {
			CopySqueakBits();
			InvalidateDeferedRegion();
		}

		PROF_ENT(8, real_time_clock_usecs() - now);
	}
}

void InvalidateDeferedRegion() {
	BRect drawRect;
	int i, count;
	
	if (DeferedRegLock.Lock()) {
		for (i = 0, count = DeferedRegion.CountRects(); i < count; i++) {
			// rectangle parameters are in squeak coordinates
			drawRect = DeferedRegion.RectAt(i);
			BMessage *message = new BMessage(DO_INVALIDATE); 
			message->AddRect("drawRect", drawRect);	
			SqWindow->PostMessage(message);
		}
		DeferedRegion.MakeEmpty();
		DeferedRegLock.Unlock();
	}
}
	
void RetrieveKey() {
	DEBUG_OUT("retrieveKey");
	if (!HasPeekKey)
		if (read_port_etc(KeyBuf, &PeekKey, 0, 0, B_TIMEOUT, 0) == 0)
			HasPeekKey = true;
	DEBUG_OUT("retrieveKey/ret");
}
	
