// this file contains: interpreter pluggable primitive support functions
//
// authors: Andreas Raab, Colin Sarsfield
// This code now bears little resemblance to Andreas' work on which it was originally
// based; however, he is still listed here to acknowledge that his code and observations
// served as starting points.
//
// Please read the file named License for license information.
// 
// To do:
// - documentation
// 

// platform-independent includes
#include "sqBeGlobal.h"

extern "C" {
	#include "sqPlatformSpecific.h"
	#include "SocketPlugin.h"
}

#include <socket.h>
#include <netdb.h>
#include <errno.h>
#include <stdexcept>

// BeOS specific includes
#include <List.h>
#include <OS.h>
#include <Locker.h>
#include <Errors.h>

// Socket types

#define TCPSocketType 0
#define UDPSocketType 1

// Resolver states

#define ResolverUninitialized 0
#define ResolverSuccess 1
#define ResolverBusy 2
#define ResolverError 3

// TCP Socket states

#define Unconnected 0
#define WaitingForConnection 1
#define Connected 2
#define OtherEndClosed 3
#define ThisEndClosed 4

// network prim global variables
static int ResolverSemaphoreIndex = 0;
static int ThisNetSession = 0;
int ResolverStatus = ResolverUninitialized;
static uint8 on = ~0;
static char LocalHostName[MAXHOSTNAMELEN + 1];
int LocalHostAddress;

int sqGetHostByAddr(int netAddress);
int sqGetHostByName(char *hostName);
int GetLocalHostAddress();

// for log message formating
char log_str[256];

// types
typedef struct {
	void *data;
	int size;
} chunk;

// async sockets vars
BList Sockets;
BLocker SocketsLock;
thread_id SocketThread;
bool LooperAlive;

class Socket;
class SocketAction;

void AddSocket(Socket *);
bool DeleteChunk(void *c);
void FinalizeSocket(Socket *s, bool close, bool destroy);
chunk *NewChunk(void *d, int s);
void SetAction(SocketAction *);
int32 SocketLooper(void *);
int SocketValid(SocketPtr s);

class SocketAction {
public:
	Socket *socket;
	SocketAction(Socket *s) { socket = s; }
	virtual bool DoIt() = 0; // if DoIt returns true then remove from action queue
};

class Socket {
public:
	int32 socket;
	int32 status;
	int32 semaIndex;
	int32 error;
	SocketAction *currentAction;
	struct sockaddr_in peer;
	
	BList sendQueue;
	BList recvQueue;	
	
	Socket(int32 s, int32 si) { socket = s; semaIndex = si; status = Unconnected; error = 0; currentAction = 0;}
	void EmptySendBuf() { sendQueue.DoForEach(DeleteChunk); sendQueue.MakeEmpty(); }
	void EmptyRecvBuf() { recvQueue.DoForEach(DeleteChunk); recvQueue.MakeEmpty(); }
	~Socket() { if (currentAction) delete currentAction; EmptySendBuf(); EmptyRecvBuf(); }
	void RecvData() {
		if (status != Connected) {
			EmptyRecvBuf();
			return;
		}
		const int max_buf = 2048;
		int buf_size;
		char rbuf[max_buf];
		try {
			while (0 < (buf_size = recv(socket, &rbuf, max_buf, 0))) {
				log_message(log_info, "sqBeNetwork: packet recevied");
				recvQueue.AddItem(NewChunk((void *)rbuf, buf_size));
				SignalSqueak();
			}
		} catch (bad_alloc) {
			log_message(log_error, "sqBeNetwork: receive data failed -- out of memory");
			return;
		}
		if (buf_size == 0) {
			if (!recvQueue.CountItems()) {
				error = ECONNRESET;
				status = OtherEndClosed;
				SignalSqueak();
			}
		} 
		else switch (errno) {
			case EINPROGRESS:
			case EWOULDBLOCK:
			case EALREADY:
				return;
			default:
				error = errno;
				SignalSqueak();
				return;
		}		
	}
	void SendData() {
		if (status != Connected) {
			EmptySendBuf();
			return;
		}
		while (sendQueue.CountItems()) {
			chunk *c = (chunk *)sendQueue.ItemAt(0);
			if (0 > send(socket, c->data, c->size, 0))
				switch (errno) {
					case ECONNRESET:
						status = OtherEndClosed;
						error = errno;
						SignalSqueak();
						return;
					case EWOULDBLOCK:
					case EINPROGRESS:
					case EALREADY:
						return;
					default:
						error = errno;
						SignalSqueak();
						return;
				}
			log_message(log_info, "sqBeNetwork: packet sent");
			DeleteChunk(c);
			sendQueue.RemoveItem((int32)0);
			SignalSqueak();
		}
	}
	void SignalSqueak() { AsyncSignalSemaphoreWithIndex(semaIndex); log_message(log_info, "sqBeNetwork: signalling squeak"); }
};

class SocketCloseAction : public SocketAction {
public:
	SocketCloseAction(Socket *s) : SocketAction(s) { }
	virtual bool DoIt() { FinalizeSocket(socket, true, false); return true; }
};

class SocketConnectAction : public SocketAction {
	int addr, port;
	bool initialized;
public:	
	SocketConnectAction(Socket *s, int a, int p) : SocketAction(s) { addr = a; port = p; initialized = false; }
	virtual bool DoIt() {
		if (!initialized) {
			initialized = true;
			memset(&socket->peer, 0, sizeof(socket->peer));
			socket->peer.sin_family = AF_INET;
			socket->peer.sin_port = htons((short)port);
			socket->peer.sin_addr.s_addr = htonl(addr);
		}
		
		int err = connect(socket->socket, (struct sockaddr *) &socket->peer, sizeof(struct sockaddr_in));
		
		if (err < 0)
			switch (errno) {
				case EWOULDBLOCK:
				case EINPROGRESS:
				case EALREADY:
					return false;
				case EISCONN:
					socket->status = Connected;		
					log_message(log_error, "sqBeNetwork: socket already connected");
					break;
				default:
					socket->error = errno;
					socket->status = Unconnected;		
					sprintf(log_str, "sqBeNetwork: socket cannot connect -- %d: %s", errno - LONG_MIN - 0x7000, strerror(errno));
					log_message(log_info, log_str);
			}
		else {
			socket->status = Connected;		
			log_message(log_info, "sqBeNetwork: socket connected");
		}
		
		socket->SignalSqueak(); 
		return true; 
	}
};

class SocketListenAction : public SocketAction {
	int port;
	bool initialized;
public:	
	SocketListenAction(Socket *s, int p) : SocketAction(s) { port = p; initialized = false; }
	virtual bool DoIt() {
		if (!initialized) {
			initialized = true;
			memset(&socket->peer, 0, sizeof(socket->peer));
			socket->peer.sin_family = AF_INET;
			socket->peer.sin_port = htons((short)port);
			socket->peer.sin_addr.s_addr = htonl(GetLocalHostAddress());
			if (0 > bind(socket->socket, (struct sockaddr *) &socket->peer, sizeof(struct sockaddr_in))
					|| 0 > listen(socket->socket, 1)) {
				socket->status = Unconnected;
				socket->error = errno;
				socket->SignalSqueak();
				log_message(log_error, "sqBeNetwork: unable to listen on socket -- listen/bind failed");
				return true;
			}		
		}
		
		if (socket->status == Unconnected || socket->status == OtherEndClosed) {
			log_message(log_info, "sqBeNetwork: socket disconnected while waiting for accept");
			return true;
		}
		
		int newSocket = accept(socket->socket, 0, NULL);
		
		if (newSocket < 0)
			switch (errno) {
				case EWOULDBLOCK:
				case EINPROGRESS:
					return false;
				default:
					socket->error = errno;
					sprintf(log_str, "sqBeNetwork: accept failed -- %s", strerror(errno));
					socket->SignalSqueak();
					return true;
			}

		closesocket(socket->socket);
		log_message(log_info, "sqBeNetwork: socket accepted");
		socket->status = Connected;		
		socket->socket = newSocket;
				
		socket->SignalSqueak(); 
		return true; 
	}
};

inline Socket * SocketByPtr(SocketPtr s) {
	return (Socket *)s->privateSocketPtr;
}

void AddSocket(Socket *socket) {
	if (SocketsLock.Lock()) {
		Sockets.AddItem((void *)socket);
		SocketsLock.Unlock();
	} else
		log_message(log_error, "sqBeNetwork: unable to add socket -- failed to obtain lock");
}

bool DeleteChunk(void *c) {
	delete[] ((chunk *)c)->data;
	delete (chunk *)c;
	return false;
}

void FinalizeSocket(Socket *s, bool close, bool destroy) {
	if (SocketsLock.Lock()) {
		int i, cnt;
		
		// delete Socket
		if (destroy) {
			Sockets.RemoveItem((void *)s);
			delete s;
		}
		
		if (close && s->status == Connected) {
			if (closesocket(s->socket) < 0)
				log_message(log_error, "sqBeNetwork: unable to abort connection -- closesocket failed");
				
			s->status = Unconnected;
		}

		SocketsLock.Unlock();
		
	} else {
		log_message(log_error, "sqBeNetwork: unable to abort connection -- failed to obtain lock");
		success(false);
	}
}

int GetLocalHostAddress() {
	struct hostent *he;
	in_addr addr; 
	
	if (!LocalHostAddress) {
		he = gethostbyname(LocalHostName);
		if (he)
			LocalHostAddress = ntohl(*(long *)(he->h_addr_list[0]));
		else
			LocalHostAddress = 0x7f000001;
		addr.s_addr = htonl(LocalHostAddress); 
	}
	return LocalHostAddress;
}
	
chunk *NewChunk(void *d, int s) {
	chunk *c = new chunk;
	c->data = (void *) new char[s];
	c->size = s;
	memcpy(c->data, d, s);
	return c;
}

void SetAction(SocketAction *action) {
	if (SocketsLock.Lock()) {
		if (action->socket->currentAction) {
			log_message(log_error, "sqBeNetwork: socket action overrun");
			delete action->socket->currentAction;
		}
		action->socket->currentAction = action;
		SocketsLock.Unlock();
	} else
		log_message(log_error, "sqBeNetwork: unable to set action -- failed to obtain lock");
}

int32 SocketLooper(void *) {
	int action, socket, actions, sockets;
	
	action = socket = 0;
	
	while (LooperAlive) 
		if (SocketsLock.Lock()) {		
		
			if (socket >= (sockets = Sockets.CountItems())) socket = 0;			
			if (sockets) {
				Socket *thisSocket = ((Socket *)Sockets.ItemAt(socket));
				
				// execute any pending action
				if (thisSocket->currentAction) {
					if (thisSocket->currentAction->DoIt())
						thisSocket->currentAction = 0;
				}

				// check for data on socket
				thisSocket->RecvData();
				// send data on socket
				thisSocket->SendData();

				socket++;
			}
			
			SocketsLock.Unlock();
			
			snooze(1000);
		}	
		
	return B_OK;
}

int SocketValid(SocketPtr s)
	{
	if ((s != NULL) 
			&& (s->privateSocketPtr != NULL) 
			&& (s->sessionID == ThisNetSession))
		return true;
	//DEBUG_OUT((stderr, "success = false(%d)\n", PSP(s)->s));
	success(false);
	return false;
	}

// Squeak network functions

// start a new network session
int sqNetworkInit(int resolverSemaIndex) {
	if (ThisNetSession != 0) return 0;  // noop if network is already initialized

	if (gethostname(LocalHostName, MAXHOSTNAMELEN) == 0)
		strcpy(LocalHostName, "localHost");
		
	LocalHostAddress = 0;
	ResolverStatus = ResolverSuccess;

	ThisNetSession = real_time_clock();
	if (ThisNetSession == 0) ThisNetSession = 1;  // don't use 0

	// install resolver semaphore
	ResolverSemaphoreIndex = resolverSemaIndex;

	// run SocketThread
	LooperAlive = true;
	SocketThread = spawn_thread(&SocketLooper, "SocketThread", B_NORMAL_PRIORITY, 0); 
		 
	if (SocketThread < B_NO_ERROR) {
		log_message(log_error, "sqBeNetwork: unable to spawn SocketLooper thread");
		success(false);
		return 0;
	}
			
	if (resume_thread(SocketThread) != B_NO_ERROR) {
		log_message(log_error, "sqBeNetwork: unable to resume SocketLooper thread");
		success(false);
		return 0;
	}

	log_message(log_info, "sqBeNetwork: network initialized");
	return 0;
}

// terminate the current network session (invalidates all open sockets)
void sqNetworkShutdown(void) {
	if (ThisNetSession == 0) return;  // noop if network is already shut down
	
	LooperAlive = false;
	
	status_t exit_value;

	if (wait_for_thread(SocketThread, &exit_value) != B_NO_ERROR)
		log_message(log_error, "sqBeNetwork: failure on network shutdown (wait for thread)");
	
	ResolverStatus = ResolverUninitialized;
	ThisNetSession = 0;
	log_message(log_info, "sqBeNetwork: network shutdown");
}

// sqSocketAbortConnection: Immediately terminate a pending connection
void sqSocketAbortConnection(SocketPtr s) {
	log_message(log_info, "sqBeNetwork: abort connection");
	FinalizeSocket(SocketByPtr(s), true, false);
}

// sqSocketCloseConnection: gracefully close an open socket
void sqSocketCloseConnection(SocketPtr s) {
	if (!SocketValid(s)) return;
	log_message(log_info, "sqBeNetwork: close connection");
	try {
		SetAction(new SocketCloseAction(SocketByPtr(s)));
	} catch (bad_alloc) {
		log_message(log_error, "sqBeNetwork: close failed -- out of memory");
	}
}

// sqSocketConnectionStatus: return public status flags of the socket
int sqSocketConnectionStatus(SocketPtr s) {
	if (!SocketValid(s)) return 0;
	return SocketByPtr(s)->status;
}

// sqSocketConnectToPort: connect() a given socket to a port
void sqSocketConnectToPort(SocketPtr s, int addr, int port) {
	if (!SocketValid(s)) return;
	log_message(log_info, "sqBeNetwork: trying to connect");
	SocketByPtr(s)->status = WaitingForConnection;
	SocketByPtr(s)->error = 0;
	try {
		SetAction(new SocketConnectAction(SocketByPtr(s), addr, port));
	} catch (bad_alloc) {
		log_message(log_error, "sqBeNetwork: connect failed -- out of memory");
	}
}

// sqSocketListenOnPort: listen() on a given port
void sqSocketListenOnPort(SocketPtr s, int port) {
	if (!SocketValid(s)) return;
	SocketByPtr(s)->status = WaitingForConnection;
	SocketByPtr(s)->error = 0;
	try {
		SetAction(new SocketListenAction(SocketByPtr(s), port));	
	} catch (bad_alloc) {
		log_message(log_error, "sqBeNetwork: listen failed -- out of memory");
	}
}

// sqSocketListenOnPortBacklogSize: listen() on a given port
void	sqSocketListenOnPortBacklogSize(SocketPtr s, int port, int backlogSize) {
	success(false);
}

// sqSocketAcceptFromRecvBytesSendBytesSemaID: accept() on a given port
void	sqSocketAcceptFromRecvBytesSendBytesSemaID(SocketPtr s, SocketPtr serverSocket, 
		int recvBufSize, int sendBufSize, int semaIndex) {
	success(false);
}

// sqSocketCreateNetTypeSocketTypeRecvBytesSendBytesSemaID:
// Create a socket for the given netType (which is always internet here)
// a given socketType (UDP or TCP) appropriate buffer size (being ignored ;-)
// and a semaphore to signal upon changes in the socket state.
void sqSocketCreateNetTypeSocketTypeRecvBytesSendBytesSemaID(SocketPtr s, int netType, 
		int socketType, int recvBufSize, int sendBufSize, int semaIndex) {
	int newSocket;
	s->sessionID = 0;
	if (socketType == 0)
		newSocket = socket(AF_INET, SOCK_STREAM, 0);
	else
		newSocket = socket(AF_INET, SOCK_DGRAM, 0);
	if (newSocket < 0) {
		log_message(log_error, "sqBeNetwork: failure to create socket");
		success(false);
		return;
	}
	
	try {
		Socket *pss = new Socket(newSocket, semaIndex);
	
		s->sessionID = ThisNetSession;
		s->socketType = socketType;
		s->privateSocketPtr = pss;
		
		setsockopt(newSocket, SOL_SOCKET, SO_NONBLOCK, &on, sizeof(on));
			
		AddSocket(pss);
		log_message(log_info, "sqBeNetwork: socket created");
	} catch (bad_alloc) {
		log_message(log_error, "sqBeNetwork: create socket failed -- out of memory");
	}
}

// sqSocketDestroy: Release the resources associated with this socket. 
// If a connection is open, it is aborted
void sqSocketDestroy(SocketPtr s) {
	if (!SocketValid(s)) return;
	log_message(log_info, "sqBeNetwork: socket destroyed");
	FinalizeSocket(SocketByPtr(s), true, true);
}

// sqSocketError: Return any error on the socket.
int sqSocketError(SocketPtr s) {
	if (!SocketValid(s)) return -1;
	
	return SocketByPtr(s)->error;
}

// sqSocketLocalAddress: Return the address of the socket on this host.
int sqSocketLocalAddress(SocketPtr s) {
	if (!SocketValid(s)) return -1;
	
	struct sockaddr_in sin;
	int sinSize = sizeof(sin);
	if (getsockname(SocketByPtr(s)->socket, (struct sockaddr *) &sin, &sinSize)) return 0;
	if (sin.sin_family != AF_INET) return 0;	
	return ntohl(sin.sin_addr.s_addr);
}

// sqSocketLocalPort: Return the port of the socket on this host
int sqSocketLocalPort(SocketPtr s) {
	if (!SocketValid(s)) return -1;

	struct sockaddr_in sin;
	int sinSize = sizeof(sin);
	if (getsockname(SocketByPtr(s)->socket, (struct sockaddr *) &sin, &sinSize)) return 0;
	if (sin.sin_family != AF_INET) return 0;	
	return ntohs(sin.sin_port);
}

// sqSocketRemoteAddress: Return the address of the socket on the remote host
int sqSocketRemoteAddress(SocketPtr s) {
	if (!SocketValid(s)) return -1;

	struct sockaddr_in sin;
	int sinSize = sizeof(sin);
	if (getpeername(SocketByPtr(s)->socket, (struct sockaddr *) &sin, &sinSize)) return 0;
	if (sin.sin_family != AF_INET) return 0;	
	return ntohl(sin.sin_addr.s_addr);
}

// sqSocketRemotePort: Return the port of the socket on the remote host
int sqSocketRemotePort(SocketPtr s) {
	if (!SocketValid(s)) return -1;

	struct sockaddr_in sin;
	int sinSize = sizeof(sin);
	if (getpeername(SocketByPtr(s)->socket, (struct sockaddr *) &sin, &sinSize)) return 0;
	if (sin.sin_family != AF_INET) return 0;	
	return ntohs(sin.sin_port);
}

// sqSocketReceiveDataAvailable: Return non-zero if data available
int sqSocketReceiveDataAvailable(SocketPtr s) {
	if (!SocketValid(s)) return 0;

	if (SocketsLock.Lock()) {
		int ret = SocketByPtr(s)->recvQueue.CountItems();		
		
		sprintf(log_str, "sqBeNetwork: receive data available -- %d packets left", ret);
		log_message(log_info, log_str);
		
		SocketsLock.Unlock();
		
		return ret;		
	} else {
		log_message(log_error, "sqBeNetwork: unable to return receive data available status -- failed to obtain lock");
		return 0;
	}
}

// sqSocketReceiveDataBufCount:
// Receive data into the given buffer. Do not exceed bufSize.
// Return the number of bytes actually read.
int sqSocketReceiveDataBufCount(SocketPtr s, int buf, int bufSize) {
	if (!SocketValid(s)) return -1;
	//if (!SocketByPtr(s)->error) return -1;

	if (SocketsLock.Lock()) {
		int ret = 0;
		char *bufptr = (char *)buf;

		while (SocketByPtr(s)->recvQueue.CountItems() && ret < bufSize) {
			chunk *c = (chunk *)SocketByPtr(s)->recvQueue.ItemAt(0);
			
			if (ret + c->size > bufSize) {
				int block = bufSize - ret;
				memcpy(bufptr, c->data, block);
				memcpy(c->data, &((char *)c->data)[block], c->size - block);
				c->size -= block;
				break;
			} else {
				memcpy(bufptr, c->data, c->size);
				ret += c->size;
				DeleteChunk(c);
				SocketByPtr(s)->recvQueue.RemoveItem((int32)0);
			}
		}
		
		sprintf(log_str, "sqBeNetwork: %d bytes received / %d packets left", ret, SocketByPtr(s)->recvQueue.CountItems());
		log_message(log_info, log_str);

		SocketsLock.Unlock();
		
		return ret;		
	} else {
		log_message(log_error, "sqBeNetwork: unable to receive data -- failed to obtain lock");
		return -1;
	}
}

// sqSocketSendDataBufCount:
// Send bufSize bytes from the data pointed to by buf. 
// Return the number of bytes sent.
int sqSocketSendDataBufCount(SocketPtr s, int buf, int bufSize) {
	if (!SocketValid(s)) return -1;

	if (SocketsLock.Lock()) {
		try {
			SocketByPtr(s)->sendQueue.AddItem(NewChunk((void *)buf, bufSize));		
		} catch (bad_alloc) {
			log_message(log_error, "sqBeNetwork: close failed -- out of memory");
			SocketsLock.Unlock();
			return -1;
		}
		
		SocketsLock.Unlock();
		
		return bufSize;		
	} else {
		log_message(log_error, "sqBeNetwork: unable to return send data -- failed to obtain lock");
		return -1;
	}
}

// sqSocketSendDone: Return non-zero if all data has been sent.
int sqSocketSendDone(SocketPtr s) {
	if (!SocketValid(s)) return 1;
	
	if (SocketsLock.Lock()) {
		int ret = !SocketByPtr(s)->sendQueue.CountItems();		
		
		sprintf(log_str, "sqBeNetwork: send done %d", ret);
		log_message(log_info, log_str);
		
		SocketsLock.Unlock();
		
		return ret;		
	} else {
		log_message(log_error, "sqBeNetwork: unable to return send done status -- failed to obtain lock");
		return 1;
	}
}



// Resolver Functions

thread_id ResolverThread = B_NO_ERROR - 1;
sem_id ResolverSem;
int LookupAddress;
char LookupName[MAXHOSTNAMELEN + 1];
int FoundAddress;
char FoundName[MAXHOSTNAMELEN + 1];
bool QuitResolver = false;
enum {lookupname, lookupaddress} LookupWhat;
int ResolverErrNo;

void DoResolverLookup();
void ResolverLookupAddress();
void ResolverLookupDone(struct hostent *he);
void ResolverLookupName();
int32 ResolverLoop(void *data);
void SignalResolverLookup();
void StartupResolverLoop();

void DoResolverLookup() {
	switch (LookupWhat) {
		case lookupname:
			ResolverLookupName();
			break;
		case lookupaddress:
			ResolverLookupAddress();
			break;
	}
}

void ResolverLookupAddress() {
	if (LookupAddress == 0x7f000001) {
		strcpy(FoundName, "localhost");
		ResolverStatus = ResolverSuccess;
		//DEBUG_OUT((stderr, "ResolverStatus = Success (ResolverLookupAddress)\n"));
		AsyncSignalSemaphoreWithIndex(ResolverSemaphoreIndex);
		return;
	}
	u_long nAddr = htonl(LookupAddress);
	ResolverLookupDone(gethostbyaddr((char *)&nAddr, 4, AF_INET));
}
	
void ResolverLookupDone(struct hostent *he) {
	if (he) {
		strcpy(FoundName, he->h_name);
		FoundAddress = ntohl(*(long *)(he->h_addr_list[0]));
		//DEBUG_OUT((stderr, "FoundAddress: %d\n", FoundAddress));
		//DEBUG_OUT((stderr, "ResolverStatus = Success (ResolverLookupDone)\n"));
		ResolverStatus = ResolverSuccess;
	} else {
		ResolverStatus = ResolverError;
		ResolverErrNo = h_errno;		
		//DEBUG_OUT((stderr, "ResolverStatus = Error (ResolverLookupDone)\n"));
	}
	AsyncSignalSemaphoreWithIndex(ResolverSemaphoreIndex);
}
	
void ResolverLookupName() {
	if (!strcmp(LookupName, "localhost")) {
		FoundAddress = 0x7f000001;
		ResolverStatus = ResolverSuccess;
		AsyncSignalSemaphoreWithIndex(ResolverSemaphoreIndex);
		//DEBUG_OUT((stderr, "ResolverStatus = Success (ResolverLookupName)\n"));
		return;
	}
	ResolverLookupDone(gethostbyname(LookupName));
}

int32 ResolverLoop(void *data) {
	while (1) {
		if (acquire_sem(ResolverSem) < B_NO_ERROR)
			fprintf(stderr, "Error acquiring resolver semaphore.\n");
		if (QuitResolver)
			return 0;
		DoResolverLookup();
	}
}
	
void SignalResolverLookup() {
	StartupResolverLoop();
	//DEBUG_OUT((stderr, "ResolverStatus = Busy (SignalResolverLookup)\n"));
	ResolverStatus = ResolverBusy;
	release_sem(ResolverSem);
}
	
void StartupResolverLoop() {
	if (ResolverThread < B_NO_ERROR) {
		ResolverThread = spawn_thread(ResolverLoop, "DNS Resolver", B_NORMAL_PRIORITY, 0);
		ResolverSem = create_sem(0, "DNS Resolver Sem");
		QuitResolver = false;
		resume_thread(ResolverThread);
	}
	
	if (ResolverThread < B_NO_ERROR)
		fprintf(stderr, "Error while spawning resolver thread.\n");	
	if (ResolverSem < B_NO_ERROR)
		fprintf(stderr, "Error while creating resolver semaphore.\n");
}


// sqResolverAbort: Abort a pending DNS lookup
void sqResolverAbort(void) {
	// is killing a thread worth it?
	while (ResolverStatus == ResolverBusy) 
		snooze(10000);	
	}

// sqResolverAddrLookupResult: Return the result of the last name lookup
void sqResolverAddrLookupResult(char *nameForAddress, int nameSize) {
	memcpy(nameForAddress, FoundName, nameSize);
}

// sqResolverAddrLookupResult: Return sizeof(result) of the last name lookup
int sqResolverAddrLookupResultSize(void) {
	return strlen(FoundName);
}

// sqResolverError: Return the last error of a DNS lookup
int sqResolverError(void) {
	if (ResolverStatus == ResolverError)
		return ResolverErrNo;
	else
		return 0;
}

// sqResolverLocalAddress: Return the address of the local host
int sqResolverLocalAddress(void) {
	return GetLocalHostAddress();
}

// sqResolverNameLookupResult: Return the address of the last DNS lookup
int sqResolverNameLookupResult(void) {
	//DEBUG_OUT((stderr, "sqResolverNameLookupResult = %d\n", FoundAddress));
	return FoundAddress;
}

// sqResolverStartAddrLookup: Look up the name to a given address
void sqResolverStartAddrLookup(int address) {
	if (ResolverStatus == ResolverBusy) return;
	if (LookupWhat == lookupaddress && LookupAddress == address) {
		// last lookup was cached :)
		AsyncSignalSemaphoreWithIndex(ResolverSemaphoreIndex);
		return;
	}
	LookupWhat = lookupaddress;
	LookupAddress = address;
	//DEBUG_OUT((stderr, "StartAddrLookup: '%8X'\n", LookupAddress));
	SignalResolverLookup();
}

// sqResolverStartNameLookup: Look up the address to a given host name
void sqResolverStartNameLookup(char *hostName, int nameSize) {
	int len;
	if (ResolverStatus == ResolverBusy) return;
	LookupWhat = lookupname;
	len = nameSize < MAXHOSTNAMELEN ? nameSize : MAXHOSTNAMELEN;
	if (LookupWhat == lookupname && strncmp(hostName, LookupName, len) == 0) {
		// last lookup was cached :)
		AsyncSignalSemaphoreWithIndex(ResolverSemaphoreIndex);
		return;
	}
	memcpy(LookupName, hostName, len);
	LookupName[len] = '\0';
	//DEBUG_OUT((stderr, "StartNameLookup: '%s'\n", LookupName));
	SignalResolverLookup();
}

// sqResolverStatus: Return resolver status
int sqResolverStatus(void) {
	//DEBUG_OUT((stderr, "ResolverStatus = %d\n", ResolverStatus));
	return ResolverStatus;	
}
	
/* module initialization/shutdown */
int socketInit(void) {
	return 1;
}

int socketShutdown(void) {
	sqNetworkShutdown();
	
	return 1;
}

void	sqSocketAcceptFromRecvBytesSendBytesSemaIDReadSemaIDWriteSemaID(
			SocketPtr s, SocketPtr serverSocket,
			int recvBufSize, int sendBufSize, int semaIndex, int readSemaIndex, int writeSemaIndex) { primitiveFail(); }
void	sqSocketCreateNetTypeSocketTypeRecvBytesSendBytesSemaIDReadSemaIDWriteSemaID(
			SocketPtr s, int netType, int socketType,
			int recvBufSize, int sendBufSize, int semaIndex, int readSemaIndex, int writeSemaIndex) { primitiveFail(); }
int 		sqSocketReceiveUDPDataBufCountaddressportmoreFlag(SocketPtr s, int buf, int bufSize,  int *address,  int *port, int *moreFlag)  { return primitiveFail(); }
int		sqSockettoHostportSendDataBufCount(SocketPtr s, int address, int port, int buf, int bufSize) { return primitiveFail(); }
int     	sqSocketSetOptionsoptionNameStartoptionNameSizeoptionValueStartoptionValueSizereturnedValue(
			SocketPtr s,int optionName, int optionNameSize, int optionValue, int optionValueSize, int *result) { return primitiveFail(); }
int     	sqSocketGetOptionsoptionNameStartoptionNameSizereturnedValue(
			SocketPtr s,int optionName, int optionNameSize, int *result) { return primitiveFail(); }

