// this file contains: BeOS serial & midi support.
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 
// To do:
// - midi input
// - refinements to midi output
// - change serial ports & midi ports to be allocated dynamically
// 

#include <MidiPort.h>
#include <MidiSynth.h>
#include <MidiText.h>
#include <SerialPort.h>
#include <OS.h>

extern "C" {
#include "sq.h"
#include "SerialPlugin.h"
#include "MIDIPlugin.h"
	}
	
//#define MIDI_DEBUG

#define MAX_SERIAL_PORTS	16

BSerialPort *serialPorts[MAX_SERIAL_PORTS] =
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

bool ValidSerialPortNum(int portNum)
	{
	return (portNum >= 0 && portNum <= MAX_SERIAL_PORTS);
	}

int serialPortInit(void) {
	return 1;
}

int serialPortShutdown(void) {
	for (int i = 0; i < MAX_SERIAL_PORTS; i++)
		serialPortClose(i);

	return 1;
}

int serialPortClose(int portNum) {
	//printf("serialPortClose\n");
	if (!ValidSerialPortNum(portNum)) { success(false); return 0; }
	
	if (serialPorts[portNum])
		{
		serialPorts[portNum]->Close();
		delete serialPorts[portNum];
		serialPorts[portNum] = 0;
		}
	
	return 0;
	}

int serialPortOpen(
  int portNum, int baudRate, int stopBitsType, int parityType, int dataBits,
  int inFlowCtrl, int outFlowCtrl, int xOnChar, int xOffChar) {
/* Open the given serial port using the given settings. The baud rate can be
   any number between about 224 and 57600; the driver will pick a clock
   divisor that will generate the closest available baud rate. */
	//printf("serialPortOpen\n");
	if (!ValidSerialPortNum(portNum)) { success(false); return 0; }
	
	serialPorts[portNum] = new BSerialPort;
	
	char portName[B_OS_NAME_LENGTH];
	
	if (B_OK != serialPorts[portNum]->GetDeviceName(portNum, portName))
		{
		fprintf(stderr, "Invalid serial port number %d.\n", portNum);
		goto no_open;
		}
	else if (0 >= serialPorts[portNum]->Open(portName))
		{
		fprintf(stderr, "Couldn't open serial port number %d.\n", portNum);		
		goto no_open;
		}
	
	serialPorts[portNum]->SetBlocking(false);

	data_rate bitsPerSecond;
	if (baudRate >= ((230400 + 115200) / 2))
		bitsPerSecond = B_230400_BPS;
	else if (baudRate >= ((57600 + 115200) / 2))
		bitsPerSecond = B_115200_BPS;
	else if (baudRate >= ((57600 + 38400) / 2))
		bitsPerSecond = B_57600_BPS;
	else if (baudRate >= ((31250 + 38400) / 2))
		bitsPerSecond = B_38400_BPS;
	else if (baudRate >= ((31250 + 19200) / 2))
		bitsPerSecond = B_31250_BPS;
	else if (baudRate >= ((9600 + 19200) / 2))
		bitsPerSecond = B_19200_BPS;
	else if (baudRate >= ((9600 + 4800) / 2))
		bitsPerSecond = B_9600_BPS;
	else if (baudRate >= ((2400 + 4800) / 2))
		bitsPerSecond = B_4800_BPS;
	else if (baudRate >= ((2400 + 1800) / 2))
		bitsPerSecond = B_2400_BPS;
	else if (baudRate >= ((1200 + 1800) / 2))
		bitsPerSecond = B_1800_BPS;
	else if (baudRate >= ((1200 + 600) / 2))
		bitsPerSecond = B_1200_BPS;
	else if (baudRate >= ((300 + 600) / 2))
		bitsPerSecond = B_600_BPS;
	else if (baudRate >= ((300 + 200) / 2))
		bitsPerSecond = B_300_BPS;
	else if (baudRate >= ((150 + 200) / 2))
		bitsPerSecond = B_200_BPS;
	else if (baudRate >= ((150 + 134) / 2))
		bitsPerSecond = B_150_BPS;
	else if (baudRate >= ((110 + 134) / 2))
		bitsPerSecond = B_134_BPS;
	else if (baudRate >= ((110 + 75) / 2))
		bitsPerSecond = B_110_BPS;
	else if (baudRate >= ((50 + 75) / 2))
		bitsPerSecond = B_75_BPS;
	else if (baudRate >= ((50 + 0) / 2))
		bitsPerSecond = B_50_BPS;
	else
		bitsPerSecond = B_0_BPS;
	serialPorts[portNum]->SetDataRate(bitsPerSecond);

	stop_bits stopBits;
	if (stopBitsType == 2)
		stopBits = B_STOP_BITS_2;
	else if (stopBitsType == 1)
		stopBits = B_STOP_BITS_1;
	else
		{
		fprintf(stderr, "Invalid stop bits on serial port number %d.\n", portNum);		
		goto no_open;
		}
	serialPorts[portNum]->SetStopBits(stopBits);
	
	parity_mode parityMode;		
	if (parityType == 0)
		parityMode = B_NO_PARITY;
	else if (parityType == 1)
		parityMode = B_ODD_PARITY;
	else if (parityType == 2)
		parityMode = B_EVEN_PARITY;
	else
		{
		fprintf(stderr, "Invalid parity mode on serial port number %d.\n", portNum);		
		goto no_open;
		}
	serialPorts[portNum]->SetParityMode(parityMode);
	
	data_bits beDataBits;
	if (dataBits == 8)
		beDataBits = B_DATA_BITS_8;
	else if (dataBits == 7)
		beDataBits = B_DATA_BITS_7;
	else
		{
		fprintf(stderr, "Invalid data bits on serial port number %d.\n", portNum);		
		goto no_open;
		}
	serialPorts[portNum]->SetDataBits(beDataBits);
		
	if (inFlowCtrl != outFlowCtrl)
		{
		fprintf(stderr, "Mismatched flow control on serial port number %d.\n", portNum);		
		goto no_open;
		}
			
	uint32 mask;
	if (inFlowCtrl == 0)
		mask = B_NOFLOW_CONTROL;
	else if (inFlowCtrl == 1)
		mask = B_SOFTWARE_CONTROL;
	else if (inFlowCtrl == 2)
		mask = B_HARDWARE_CONTROL;
	else
		{
		fprintf(stderr, "Invalid flow control on serial port number %d.\n", portNum);		
		goto no_open;
		}
	serialPorts[portNum]->SetFlowControl(mask);

	if (xOnChar != 19 || xOffChar != 24)
		{
		fprintf(stderr, "Cannot change XON/XOFF characters on serial port number %d.\n", portNum);		
		goto no_open;
		}
	return 0;
	
no_open:
	serialPortClose(portNum);
	success(false);
	return 0;
	}

int serialPortMidiClockRate(int portNum, int interfaceClockRate) {
/* Put the given port into MIDI mode, which uses a clock supplied
   by an external MIDI interface adaptor to determine the baud rate.
   Possible external clock rates: 31.25 KHz, 0.5 MHz, 1 MHz, or 2 MHz. */
	//printf("serialPortMidiClockRate\n");
	success(false);
	return 0;
	}

int serialPortReadInto(int portNum, int count, int startPtr) {
/* Read up to count bytes from the given serial port into the given byte array.
   Read only up to the number of bytes in the port's input buffer; if fewer bytes
   than count have been received, do not wait for additional data to arrive.
   Return zero if no data is available. */
	//printf("serialPortReadInto\n");
	if (!ValidSerialPortNum(portNum)) { success(false); return 0; }
	if (!serialPorts[portNum]) { success(false); return 0; }
	
	int r = serialPorts[portNum]->Read((void *)startPtr, count);	
	if (r < 0)
		{
		success(false);
		return 0;
		}
	else
		return r;
	}

int serialPortWriteFrom(int portNum, int count, int startPtr) {
/* Write count bytes from the given byte array to the given serial port's
   output buffer. Return the number of bytes written. This implementation is
   synchronous: it doesn't return until the data has been sent. However, other
   implementations may return before transmission is complete. */
	//printf("serialPortWriteFrom\n");
	if (!ValidSerialPortNum(portNum)) { success(false); return 0; }
	if (!serialPorts[portNum]) { success(false); return 0; }

	int r = serialPorts[portNum]->Write((void *)startPtr, count);	
	if (r < 0)
		{
		success(false);
		return 0;
		}
	else
		return r;
	}

class SqueakMidiOut: public BMidi
	{
public:
	port_id data_port;
	
	SqueakMidiOut() 
		{ data_port = create_port(1024, "Midi out port"); }
	~SqueakMidiOut() 
		{ delete_port(data_port); }
	void Run() 
		{ 
		uint32 atime;
		int nbytes;
		uchar channel;
		uchar buf[20];	
		while (KeepRunning()) 
			{ 
			if ((nbytes = read_port_etc(data_port, (int32 *)&atime, buf, 20, B_TIMEOUT, 10000)) >= 0)
				{
				channel = (buf[0] & 0xf) + 1;
				switch (buf[0] & 0xf0)
					{
					case B_NOTE_ON:
		      			SprayNoteOn(channel, buf[1], buf[2], atime);
		      			break;				
					case B_NOTE_OFF:
		      			SprayNoteOff(channel, buf[1], buf[2], atime);
		      			break;				
					case B_PROGRAM_CHANGE:
      					SprayProgramChange(channel, buf[1], atime);
      					break;				
					case B_CONTROL_CHANGE:
      					SprayControlChange(channel, buf[1], buf[2], atime);
      					break;				
					case B_KEY_PRESSURE:
      					SprayKeyPressure(channel, buf[1], buf[2], atime);
      					break;				
					case B_CHANNEL_PRESSURE:
      					SprayChannelPressure(channel, buf[1], atime);
      					break;				
					case B_PITCH_BEND:
      					SprayPitchBend(channel, buf[1], buf[2], atime);
      					break;
		      		case B_SYS_EX_START:				
        				/* Meta or Sysex command */
						SpraySystemExclusive((void *)buf, nbytes, atime);
						break;
					default:
						fprintf(stderr, "unrecognized midi command %x\n", (int)buf[0]);
					}
				}
			} 
		}
	};

#define MAX_MIDI_PORTS 4

BMidiPort *midiPorts[MAX_MIDI_PORTS];
BMidiSynth *theSynth;
SqueakMidiOut *squeakOuts[MAX_MIDI_PORTS + 1] = {0, 0, 0, 0, 0};
int portCount = -1;

#if defined(MIDI_DEBUG)
	// this var is for debugging only
	BMidiText theText;
#endif

bool validMidiPortNum(int portNum)
	{
	return (portNum >= 0 && portNum <= portCount);
	}

bool openMidiPortNum(int portNum)
	{
	if (portNum >= 0 && portNum <= portCount)
		return squeakOuts[portNum] != 0;
	else
		return false;
	}

int midiInit(void) {
	return 1;
}

int midiShutdown(void) {
	for (int i = 0; i < portCount; i++)
		sqMIDIClosePort(i);

	return 1;
}

int sqMIDIClosePort(int portNum) {
/* Close the given MIDI port. Do nothing if the port is not open.
   Fail if there is no port of the given number.*/
	//printf("sqMIDIClosePort\n");
	if (!validMidiPortNum(portNum)) { success(false); return 0; }	
	if (!openMidiPortNum(portNum)) { return 0; }
	
	squeakOuts[portNum]->Stop();
	
	if (portNum)
		{
		squeakOuts[portNum]->Disconnect(midiPorts[portNum-1]);
		midiPorts[portNum - 1]->Close();
		delete midiPorts[portNum - 1];
		}
	else
		{
		#if defined(MIDI_DEBUG)
			squeakOuts[portNum]->Disconnect(&theText);
		#else
			squeakOuts[portNum]->Disconnect(theSynth);
		#endif
		theSynth->EnableInput(false, true);
		
		delete theSynth;
		}
		
	delete squeakOuts[portNum];
	squeakOuts[portNum] = 0;
		
	return 0;
} 

int sqMIDIGetClock(void) {
/* Return the current value of the clock used to schedule MIDI events.
   The MIDI clock is assumed to wrap at or before half the maximum
   positive SmallInteger value. This allows events to be scheduled
   into the future without overflowing into LargePositiveIntegers. 
   This implementation does not support event scheduling, so it
   just returns the value of the Squeak millisecond clock. */
	uint32 t = (system_time() / 1000) & 0x3fffffff;
	//printf("sqMIDIGetClock %d\n", t);
	return t;
}

int sqMIDIGetPortCount(void) {
/* Return the number of available MIDI interfaces, including both
   hardware ports and software entities that act like ports. Ports
   are numbered from 0 to N-1, where N is the number returned by this
   primitive. */
   
	BMidiPort port;
   
   	portCount = min_c(port.CountDevices(), MAX_MIDI_PORTS);
   	
	return portCount + 1;
}

int sqMIDIGetPortDirectionality(int portNum) {
/* Return an integer indicating the directionality of the given
   port where: 1 = input, 2 = output, 3 = bidirectional. Fail if
   there is no port of the given number. */

	if (!validMidiPortNum(portNum)) { success(false); return 0; }
	if (portNum)
		return 2; //return 3; will be 3 when we get input working
	else
		return 2;
}

int sqMIDIGetPortName(int portNum, int namePtr, int length) {
/* Copy the name of the given MIDI port into the string at the given
   address. Copy at most length characters, and return the number of
   characters copied. Fail if there is no port of the given number.*/
  
	//printf("sqMIDIGetPortName\n");
	if (!validMidiPortNum(portNum)) { success(false); return 0; }
	if (portNum)
		{
		BMidiPort port;
		status_t ret;
		char name[B_OS_NAME_LENGTH];
		int len = 0;
	
		ret = port.GetDeviceName(portNum - 1, name);
		if (ret != B_OK)
			success(false);
		else
			{
			len = min_c(length, (int)strlen(name));
			memcpy((void *)namePtr, name, len);
			}
		return len;		
		}
	else
		{
		const char *synthName = "Headspace software synthsizer";
		memcpy((void *)namePtr, synthName, strlen(synthName));
	
		return strlen(synthName);
		}
	}

int sqMIDIOpenPort(int portNum, int readSemaIndex, int interfaceClockRate) {
/* Open the given port, if possible. If non-zero, readSemaphoreIndex
   specifies the index in the external objects array of a semaphore
   to be signalled when incoming MIDI data is available. Note that
   not all implementations support read semaphores (this one does
   not); see sqMIDICanUseSemaphore. The interfaceClockRate parameter
   specifies the clock speed for an external MIDI interface
   adaptor on platforms that use such adaptors (e.g., Macintosh).
   Fail if there is no port of the given number.*/
	//printf("sqMIDIOpenPort\n");
	if (!validMidiPortNum(portNum)) { success(false); return 0; }
	
	sqMIDIClosePort(portNum);

	squeakOuts[portNum] = new SqueakMidiOut;
	
	if (portNum)
		{
		char name[B_OS_NAME_LENGTH];
	
		midiPorts[portNum - 1] = new BMidiPort;
		if (midiPorts[portNum - 1]->GetDeviceName(portNum - 1, name) != B_OK)
			goto no_midi_open;
		else if (midiPorts[portNum - 1]->Open(name) == B_OK)
			squeakOuts[portNum]->Connect(midiPorts[portNum-1]);
		else
			goto no_midi_open;
		}
	else
		{
		theSynth = new BMidiSynth;
		theSynth->EnableInput(true, true);
		#if defined(MIDI_DEBUG)
			squeakOuts[portNum]->Connect(&theText);
		#else
			squeakOuts[portNum]->Connect(theSynth);
		#endif
		}
		
	squeakOuts[portNum]->Start();
	
	return 0;
no_midi_open:
	sqMIDIClosePort(portNum);
	success(false);
	return 0;
}

/*** MIDI Parameters (used with sqMIDIParameter function) ***/

#define sqMIDIInstalled				1
/* Read-only. Return 1 if a MIDI driver is installed, 0 if not.
   On OMS-based MIDI drivers, this returns 1 only if the OMS
   system is properly installed and configured. */

#define sqMIDIVersion				2
/* Read-only. Return the integer version number of this MIDI driver.
   The version numbering sequence is relative to a particular driver.
   That is, version 3 of the Macintosh MIDI driver is not necessarily
   related to version 3 of the Win95 MIDI driver. */

#define sqMIDIHasBuffer				3
/* Read-only. Return 1 if this MIDI driver has a time-stamped output
   buffer, 0 otherwise. Such a buffer allows the client to schedule
   MIDI output packets to be sent later. This can allow more precise
   timing, since the driver uses timer interrupts to send the data
   at the right time even if the processor is in the midst of a
   long-running Squeak primitive or is running some other application
   or system task. */

#define sqMIDIHasDurs				4
/* Read-only. Return 1 if this MIDI driver supports an extended
   primitive for note-playing that includes the note duration and
   schedules both the note-on and the note-off messages in the
   driver. Otherwise, return 0. */

#define sqMIDICanSetClock			5
/* Read-only. Return 1 if this MIDI drivers clock can be set
   via an extended primitive, 0 if not. */

#define sqMIDICanUseSemaphore		6
/* Read-only. Return 1 if this MIDI driver can signal a semaphore
   when MIDI input arrives. Otherwise, return 0. If this driver
   supports controller caching and it is enabled, then incoming
   controller messages will not signal the semaphore. */

#define sqMIDIEchoOn				7
/* Read-write. If this flag is set to a non-zero value, and if
   the driver supports echoing, then incoming MIDI events will
   be echoed immediately. If this driver does not support echoing,
   then queries of this parameter will always return 0 and
   attempts to change its value will do nothing. */

#define sqMIDIUseControllerCache	8
/* Read-write. If this flag is set to a non-zero value, and if
   the driver supports a controller cache, then the driver will
   maintain a cache of the latest value seen for each MIDI controller,
   and control update messages will be filtered out of the incoming
   MIDI stream. An extended MIDI primitive allows the client to
   poll the driver for the current value of each controller. If
   this driver does not support a controller cache, then queries
   of this parameter will always return 0 and attempts to change
   its value will do nothing. */

#define sqMIDIEventsAvailable		9
/* Read-only. Return the number of MIDI packets in the input queue. */

#define sqMIDIFlushDriver			10
/* Write-only. Setting this parameter to any value forces the driver
   to flush its I/0 buffer, discarding all unprocessed data. Reading
   this parameter returns 0. Setting this parameter will do nothing
   if the driver does not support buffer flushing. */

#define sqMIDIClockTicksPerSec		11
/* Read-only. Return the MIDI clock rate in ticks per second. */

#define sqMIDIHasInputClock			12
/* Read-only. Return 1 if this MIDI driver timestamps incoming
   MIDI data with the current value of the MIDI clock, 0 otherwise.
   If the driver does not support such timestamping, then the
   client must read input data frequently and provide its own
   timestamping. */

int sqMIDIParameter(int whichParameter, int modify, int newValue) {
/* Read or write the given MIDI driver parameter. If modify is 0,
   then newValue is ignored and the current value of the specified
   parameter is returned. If modify is non-zero, then the specified
   parameter is set to newValue. Note that many MIDI driver parameters
   are read-only; attempting to set one of these parameters fails.
   For boolean parameters, true = 1, false = 0. */
	//printf("sqMIDIParameter\n");

	if (modify == 0) {
		switch(whichParameter) {
		case sqMIDIVersion:
			return 1;
			break;
		case sqMIDIInstalled:
		case sqMIDIHasBuffer:
			return 1;
			break;
		case sqMIDIHasDurs:
		case sqMIDICanSetClock:
		case sqMIDICanUseSemaphore:
		case sqMIDIEchoOn:
		case sqMIDIUseControllerCache:
		case sqMIDIEventsAvailable:
		case sqMIDIFlushDriver:
		case sqMIDIHasInputClock:
			return 0;
			break;
		case sqMIDIClockTicksPerSec:
			return 1000;
			break;
		default:
			return success(false);
		}
	} else {
		switch(whichParameter) {
		case sqMIDIInstalled:
		case sqMIDIVersion:
		case sqMIDIHasBuffer:
		case sqMIDIHasDurs:
		case sqMIDICanSetClock:
		case sqMIDICanUseSemaphore:
			return success(false);
			break;
		case sqMIDIEchoOn:
			/* noop; echoing not supported */
			break;
		case sqMIDIUseControllerCache:
			/* noop; controller cache not supported */
			break;
		case sqMIDIEventsAvailable:
			return success(false);
			break;
		case sqMIDIFlushDriver:
			/* noop; buffer flushing not supported */
			break;
		case sqMIDIClockTicksPerSec:
			return success(false);
			break;
		default:
			return success(false);
		}
	}
	return 0;
}

int sqMIDIParameterSet(int whichParameter, int newValue) {
	success(false);
}

int sqMIDIParameterGet(int whichParameter) {
	success(false);
}

int sqMIDIPortReadInto(int portNum, int count, int bufferPtr) {
/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Copy up to (length - 4) bytes
   of incoming MIDI data into that buffer, preceded by a 4-byte
   timestamp in the units of the MIDI clock, most significant byte
   first. Implementations that do not support timestamping of
   incoming data as it arrives (see sqMIDIHasInputClock) simply
   set the timestamp to the value of the MIDI clock when this
   function is called. Return the total number of bytes read,
   including the timestamp bytes. Return zero if no data is
   available. Fail if the buffer is shorter than five bytes,
   since there must be enough room for the timestamp plus at
   least one data byte. */
	//printf("sqMIDIPortReadInto\n");
	if (!validMidiPortNum(portNum)) { success(false); return 0; }
  success(false);
  return 0;
}

int sqMIDIPortWriteFromAt(int portNum, int count, int bufferPtr, int time) {
/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Send its contents to the given
   port when the MIDI clock reaches the given time. If time equals
   zero, then send the data immediately. Implementations that do
   not support a timestamped output queue, such as this one, always
   send the data immediately; see sqMIDIHasBuffer. */
	//printf("sqMIDIPortWriteFromAt\n");
   
	if (!openMidiPortNum(portNum)) { success(false); return 0; }
	
	uint32 atime;
	if (time != 0)
		atime = time + system_time() / 1000;
	else
		atime = B_NOW;
		
	write_port(squeakOuts[portNum]->data_port, atime, (void *)bufferPtr, count);	
	
	return count;
	}
