// this file contains: code that is shared between the BWindow implementation
// and the BDirectWindow implementation
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 

#include "sqBeSharedWindow.h"
#include "sqBeView.h"
#include "sqBeApp.h"
#include "sqBeIOPrims.h"
#include "sqBeGlobal.h"
#include "sqBeLogWindow.h"
#include <MenuBar.h>
#include <MenuItem.h>
#include <Screen.h>
#include <Beep.h>
#include <Alert.h>

#define SQ_DEFER_UPDATES 'sqDU'
#define SQ_SHOW_LOG_WINDOW 'sqLW'

char ConvertChar(char beChar);

int32 UsedTopSpace = 0;

void CreateViews(BWindow *win, BRect &bounds)
	{
	BRect r;
	
 	win->Lock();

	r = win->Bounds();
	r.bottom = r.top + 14;

	BMenuBar *mb = new BMenuBar(r,"main menu");
	BMenu *m;
	BMenuItem *i;

	mb->AddItem(m = new BMenu("File"));

	m->AddItem(i = new BMenuItem("About BeSqueak", new BMessage(B_ABOUT_REQUESTED)));
	i->SetTarget(be_app);	
	m->AddSeparatorItem();
	m->AddItem(i = new BMenuItem("Quit", new BMessage(B_QUIT_REQUESTED)));
	i->SetTarget(be_app);	

	mb->AddItem(m = new BMenu("Preferences"));
	m->AddItem(i = new BMenuItem("Defer Updates", new BMessage(SQ_DEFER_UPDATES)));
	i->SetMarked(fDeferredUpdate);
	m->AddItem(LogWindowItem = new BMenuItem("Show Log Window", new BMessage(SQ_SHOW_LOG_WINDOW)));	

	win->AddChild(mb);
	mb->ResizeToPreferred();
	
	r = mb->Bounds();
	UsedTopSpace = r.IntegerHeight() + 1;

	r = bounds;
	r.top = mb->Bounds().bottom+1;
	SqView = new SqueakView(r);
	win->AddChild(SqView);
	win->SetPulseRate(55000);
	win->Unlock();

	}

bool GetSqueakDisplayInfo(int &depth, int &width, int &height, int &dispBitsIndex)
	{
	int dispBits, displayObj;
	displayObj = displayObject();
	if ((isPointers(displayObj)) && ((lengthOf(displayObj)) >= 4)) 
		{
		dispBits = fetchPointerofObject(0, displayObj);
		width = fetchIntegerofObject(1, displayObj);
		height = fetchIntegerofObject(2, displayObj);
		depth = fetchIntegerofObject(3, displayObj);
		dispBitsIndex = dispBits + 4;
		return false;
		}
	return true;
	}

// returns true if inherited DispatchMessage should be called
bool HandleDispatchMessage(BWindow *win, BMessage *msg, BHandler *target) {
	switch ( msg->what ) { 
		case B_KEY_DOWN: 
		case B_KEY_UP: 
		{
			int keystate;
			int8 byte;
			int64 when;
    		int32 mods;// = modifiers();
    		int32 repeat = 0;
    		win->CurrentMessage()->FindInt32("modifiers", &mods);
    		win->CurrentMessage()->FindInt64("when", &when);
    		win->CurrentMessage()->FindInt32("repeat", &repeat);
    		
	   		ConvertModifiers(mods);
			msg->FindInt8("byte", &byte);
			keystate = ((ButtonState & (~7)) << 5) | ConvertChar(byte);
			if (keystate == interruptKeycode && msg->what == B_KEY_DOWN) {
				/* Note: interrupt key is "meta"; it not reported as a keystroke */
				interruptPending = true;
				interruptCheckCounter = 0;
			} 
			else if (SqEvents) {
				sqKeyboardEvent evt;
				evt.type = EventTypeKeyboard;
				evt.charCode = ConvertChar(byte);
				evt.pressCode = msg->what == B_KEY_DOWN ? EventKeyDown : EventKeyUp;
		   		evt.modifiers = ((mods & B_COMMAND_KEY) ? CommandKeyBit : 0)
					| ((mods & B_OPTION_KEY) ? OptionKeyBit : 0)
					| ((mods & B_CONTROL_KEY) ? CtrlKeyBit : 0)
					| ((mods & (B_SHIFT_KEY | B_CAPS_LOCK)) ? ShiftKeyBit : 0);
		   		evt.timeStamp = (int)(when / 1000);
		   		if (repeat == 0) {
					if (write_port_etc(SqEvents, 0, (void*)&evt, sizeof(evt), B_TIMEOUT, 0) != B_NO_ERROR)
						beep();
		   		}
		   		if (msg->what == B_KEY_DOWN) {
					evt.pressCode = EventKeyChar;
					if (write_port_etc(SqEvents, 0, (void*)&evt, sizeof(evt), B_TIMEOUT, 0) != B_NO_ERROR)
						beep();
		   		}
		   		
				AsyncSignalSemaphoreWithIndex(SqEventSem);
			} else if (msg->what == B_KEY_DOWN)
				if (write_port_etc(KeyBuf, keystate, 0, 0, B_TIMEOUT, 0) != B_NO_ERROR)
					beep();
				
			return false; 
		}
		case SQ_DEFER_UPDATES: {
			//msg->PrintToStream();
			BMenuItem *i;
			if (msg->FindPointer("source", (void **)&i) == B_OK) {
				i->SetMarked(fDeferredUpdate = !fDeferredUpdate);
				SqView->Invalidate();
			}
			return false;
		}
		case SQ_SHOW_LOG_WINDOW:
			LogWindowSetVisibleTo(!ShowLogWindow);
			return false;
		default: 
     		return true; 
	} 
}

void HandleFrameResized(BWindow *win, float width, float height) {
	BScreen screen(win);
	if (width != screen.Frame().Width() || height != screen.Frame().Height()) {
		savedWindowSize = (((uint32)width + 1) << 16) | ((uint32)height + 1);
		//addEventToBuffer(1);
	}
}		

void HandleQuitRequested()
	{
	if (!Quitting)
		{
		BAlert *alert = new BAlert("Confirmation", 
				"Exit Squeak without saving?", 
				"Yes", "No", NULL, B_WIDTH_FROM_LABEL, B_STOP_ALERT); 
		if (alert->Go() == 1)
			return;
		}
	#if defined(DEBUG_SQUEAK)
		printCallStack();
	#endif
	
	Quitting = true;

	// wait for interpreter thread to die, if it doesn't by the time timeout
	// is reached, then kill it
	thread_info interp_info;
	const bigtime_t timeout = 2000000; // 2 seconds
	bigtime_t start = system_time();	
	status_t result = B_NO_ERROR;
	
	while (system_time() - start < timeout && result == B_NO_ERROR) {
		result = get_thread_info(InterpretThread, &interp_info);
		snooze(100000);
	}

	if (result == B_NO_ERROR)
		{
		fprintf(stderr, "Killing interpreter thread...\n");
		kill_thread(InterpretThread);
		}
	
	BMessage msg(B_QUIT_REQUESTED);
	SqApp->PostMessage(&msg);
	}

// support functions

char ConvertChar(char beChar)
   	{
   	switch (beChar)
   		{
   		case 10:
   			return 13; 
   		default:
   			return beChar;
   		}
   	}
   	
void ConvertModifiers(uint32 modifiers)
	{ 
	DEBUG_OUT("ConvertModifiers");
	ButtonState &= 7;
	// pass key info into structure
	ButtonState |= ((modifiers & B_COMMAND_KEY) ? 64 : 0)
		| ((modifiers & B_OPTION_KEY) ? 32 : 0)
		| ((modifiers & B_CONTROL_KEY) ? 16 : 0)
		| ((modifiers & (B_SHIFT_KEY | B_CAPS_LOCK)) ? 8 : 0);
	//printf("buttonState: %d\n", buttonState);
	DEBUG_OUT("ConvertModifiers/ret");
	}
	
