// this file contains: BeOS sound support.
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 
// This file is a rewrite of the file "sqMacSound.c".
//
// The original version of this file can be regenerated from the Squeak
// image by executing:
//
// 	InterpreterSupportCode writeMacSourceFiles
//
// Currently the default is three buffers which seems to work pretty
// well.  Note that a minimum of three buffers is required to keep
// the insert primitive from returning too small a number which causes
// Squeak to get an index out of bounds error on occasion.
//
// Note that these calculations always assume that Squeak is requesting
// stereo output. 
//

#include <MediaDefs.h>
#include <SoundPlayer.h>
#include <ByteOrder.h>
#include <OS.h>
#include <cstdio>

extern "C" {
	#include "SoundPlugin.h"
	#include "sq.h"
	}

#define FAIL(X) { return X; }

//#define SOUND_DEBUG_MSGS

// See the Be Book for details on BSoundPlayer.  Note that when this
// was written the constructor I used was not yet documented in the Be Book
// it was however mentioned off-handedly in a newsletter article.

void BufferProc(void *theCookie, void *buffer, size_t size, 
   		const media_raw_audio_format &format);
   		
class SqueakSound
	{ 
	BSoundPlayer *SoundPlayer;
public: 
	SqueakSound() 
		{ 
		IsPlaying = false; 
		}
	~SqueakSound()
		{
		Stop();
		}
	int AvailableSpace(void);
	int InsertSamplesFromLeadTime(int frameCount, int srcBufPtr, int samplesOfLeadTime);
	int PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex);
	int PlaySilence(void);
	int Start(int frameCount, int samplesPerSec, int stereo, int semaIndex); 
	void Stop(void);
	
	bool IsPlaying, stereo;
	int FrameCount;
	int BufferSize;
	int SemaIndex;
	
	static const int max_sound_bufs = 3; // see note at top of file before changing

	enum buf_state {full, empty, playing};

	struct {
		int16 *Buffer, *End;
		volatile int16 *Position;
		volatile enum buf_state BufState;
		bigtime_t fill_time;
		} SoundBuf[max_sound_bufs];

	bool FillBuffer(void *buffer, size_t size, 
    		const media_raw_audio_format &format); 
private: 
	int GetEarliestBufWithState(buf_state bs, bigtime_t later_than = 0);
	int GrabBuf();
	};

int SqueakSound::AvailableSpace()
	// return -1 for failure, return 0 if all buffers full,
	// return size of buffer if space is available
	{
	int i; 
	if (!IsPlaying) return -1;
	for (i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == empty)
			return BufferSize;
	return 0;	
	}

void BufferProc(void *theCookie, void *buffer, size_t size, 
    		const media_raw_audio_format &format)
    {
    ((SqueakSound *)theCookie)->FillBuffer(buffer, size, format);
    }
    		
int SqueakSound::GetEarliestBufWithState(buf_state bs, bigtime_t later_than)
	{
	bigtime_t mintime = B_INFINITE_TIMEOUT;
	int i, min = -1;
	for (i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == bs && SoundBuf[i].fill_time > later_than && SoundBuf[i].fill_time < mintime)
			{
			mintime = SoundBuf[i].fill_time;
			min = i;
			}
			
	return min;
	}
	
int SqueakSound::GrabBuf()
	// grab a buffer for playing for use by the FillBuffer function.
	// return buffer index if successful, -1 if not
	{
	int i;
	for (i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == playing)
			return i;
			
	i = GetEarliestBufWithState(full);
	if (i != -1)
		{
		SoundBuf[i].BufState = playing;
		return i;
		}

	return -1;
	}
	
int SqueakSound::InsertSamplesFromLeadTime(int frameCount, int srcBufPtr, int samplesOfLeadTime)
	// this function performs a sound overlay on buffered data, even while it is playing
	// notice that samplesOfLeadTime indicates how many frames ahead of the current spot
	// that is playing you want to be.
	// This function returns -1 if the sound is off, or the number of frames successfully
	// overlaid/played if successful.  Note that if this function returns < the size of
	// one buffer, Squeak often gets very unhappy.
	{
	if (!IsPlaying) return -1;

	int i;
	int Samples = 0;
	int16 *src, *src_end, *dst_end, *dst;

	src = (int16 *) srcBufPtr;
	src_end = (int16 *) (srcBufPtr + (frameCount << 2));

	for (i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == playing)
			break;
	
	if (i != max_sound_bufs)
		{
	    dst_end = SoundBuf[i].End;
		dst = (int16 *) (SoundBuf[i].Position + (samplesOfLeadTime << 1));		      
	
		Samples = ((int)dst_end - (int)dst) >> 2;

		while (src < src_end && dst < dst_end) *dst++ += *src++;
		}

	int framesWritten = Samples;
	
	if (framesWritten < 0) framesWritten = 0;

	bigtime_t last_buf_time;
	
	i = GetEarliestBufWithState(full);

	while (i != -1)
		{
	    dst_end = SoundBuf[i].End;
		dst = SoundBuf[i].Buffer;		      
	    if (Samples < 0)
	    	dst += (-Samples) << 1;
		
		Samples = min_c(((int)dst_end - (int)dst) >> 2,
				((int)src_end - (int)src) >> 2);
	
		while (src < src_end && dst < dst_end) *dst++ += *src++;
			
		if (Samples < 0) Samples = 0;
		
		framesWritten += Samples;
	
		last_buf_time = SoundBuf[i].fill_time;
		i = GetEarliestBufWithState(full, last_buf_time);
		}

	// if you can't overlay them, then play them. ;)
	if (framesWritten < frameCount)
		framesWritten += PlaySamplesFromAtLength(frameCount - framesWritten, srcBufPtr, framesWritten);
	
	// this takes less clock cycles then doing it the right way,
	// and gets the same answer
	if (framesWritten > frameCount) framesWritten = frameCount;
	
	return framesWritten;
	}
	
bool SqueakSound::FillBuffer(void *buffer, size_t size, 
    		const media_raw_audio_format &format)
	{ 
	int16 *soundData;		// Pointer to the frame to alter 
	int32 frameCount;		// Number of frames to manage 
	
	int curBuf;
	//printf("sampling rate: %f, channel_count: %ld, format: %ld, byte_order: %ld, buffer_size: %ld\n",
	//	format.frame_rate, format.channel_count, format.format, format.byte_order, format.buffer_size);
	
    //return true;
      
    curBuf = GrabBuf();
    if (curBuf == -1) return true;
    
	soundData = (int16 *) buffer;       
	frameCount = size >> 2; 
      
	while (frameCount-- > 0)
		{
		*soundData++ = *SoundBuf[curBuf].Position++;
		*soundData++ = *SoundBuf[curBuf].Position++;
		if (SoundBuf[curBuf].Position >= SoundBuf[curBuf].End)
			{
			SoundBuf[curBuf].BufState = empty;
			AsyncSignalSemaphoreWithIndex(SemaIndex);
			curBuf = GrabBuf();
			if (curBuf == -1) break;
			}
		} 
     
	return true; 
	}

int SqueakSound::PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
	{
	int framesWritten;

	if (!IsPlaying) return -1;

	int i;
	for (i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == empty)
			break;
			
	if (i == max_sound_bufs)
		return 0;

	if (FrameCount < frameCount)
		framesWritten = FrameCount;
	else
		framesWritten = frameCount;

	int16 *src = (int16 *) (arrayIndex + (startIndex << 2)),
	      *end = (int16 *) (arrayIndex + ((startIndex + framesWritten) << 2)),
	      *dst = (int16 *) SoundBuf[i].Buffer;
	
	while (src < end) *dst++ = *src++;

	SoundBuf[i].Position = SoundBuf[i].Buffer;
	SoundBuf[i].fill_time = system_time();
	SoundBuf[i].BufState = full;
	
	return framesWritten;
	}     
	
int SqueakSound::PlaySilence(void)
	{
	if (!IsPlaying) return -1;

	for (int i = 0; i < max_sound_bufs; i++)
		if (SoundBuf[i].BufState == empty)
			{
			int16 *buf = SoundBuf[i].Buffer;
			int16 *bufEnd = (int16 *)&((char *)buf)[BufferSize];
			while (buf < bufEnd) *buf++ = 0;
			SoundBuf[i].BufState = full;
			return BufferSize;
			}
	return 0;	
	}
     
int SqueakSound::Start(int frameCount, int samplesPerSec, int stereo, int semaIndex)
	// return 1 if success, 0 if failure
	{ 
	media_raw_audio_format format;
	// stop if already playing
	if (IsPlaying) Stop();
	
	if (!stereo)
		{
		printf("sqBeSound.cpp: Unable to support request for mono (single channel) sound.\n");
		return 0;
		}
	format.frame_rate = ((float)samplesPerSec);
	format.channel_count = stereo?2:1;
	format.format = media_raw_audio_format::B_AUDIO_SHORT;
	#if B_HOST_IS_BENDIAN
	format.byte_order = B_MEDIA_BIG_ENDIAN;
	#else
	format.byte_order = B_MEDIA_LITTLE_ENDIAN;
	#endif
	format.buffer_size = frameCount;

	try {
		SoundPlayer = new BSoundPlayer(&format, "Squeak Sound Output", &BufferProc, NULL, (void *)this);
	} catch(...) {
		printf("sqBeSound.cpp: Failed to create BSoundPlayer.\n");
		return 0;
	}

	// size of sound buffer
	FrameCount = frameCount;
	
	// these seem to always be true as far as the Squeak format goes
	// but my investigating has been light so far.
	this->stereo = true;  // changing this could also break some code.
	BufferSize = 4 * FrameCount;
	SemaIndex = semaIndex;  // this semaphore signals Squeak that a buffer
	                        // is now empty
	
	// alloc & init the sound buffers
	for (int i = 0; i < max_sound_bufs; i++)
		{
		SoundBuf[i].Buffer = (int16 *)malloc(4 * FrameCount);
		SoundBuf[i].End = (int16 *)&((char *)SoundBuf[i].Buffer)[BufferSize];
		SoundBuf[i].BufState = empty;
		if (SoundBuf[i].Buffer == NULL)
			{
			printf("sqBeSound.cpp: Unable to allocate memory for sound buffer.\n");
			return 0;
			}
		}
		
	SoundPlayer->Start();
	SoundPlayer->SetHasData(true);
	SoundPlayer->SetVolume(0.725);
	
	// we are now playing
	IsPlaying = true;
	return 1;
	}

void SqueakSound::Stop(void)
	{ 
	// no op if not playing
	if (IsPlaying)
		{
		// no longer playing
		IsPlaying = false;	

		SoundPlayer->Stop();
		delete SoundPlayer; 
		}
	}

// the Squeak Sound output object

SqueakSound SqSound;

/* module initialization/shutdown */
int soundInit(void)
	{
#ifdef SOUND_DEBUG_MSGS
	printf("soundInit()\n");
#endif
	return 1;
	}
	
int soundShutdown(void)
	{
#ifdef SOUND_DEBUG_MSGS
	printf("soundShutdown()\n");
#endif
	return 1;
	}

// sound output

void snd_Volume(double *left, double *right) {
	// ATTN:010309:Implement
}

void snd_SetVolume(double left, double right) {
	// ATTN:010309:Implement
}

int snd_AvailableSpace(void)
	{    
	int result;
	result = SqSound.AvailableSpace();
#ifdef SOUND_DEBUG_MSGS
	printf("%d = snd_AvailableSpace()\n", result);
#endif
	return result;
    }
    
int snd_InsertSamplesFromLeadTime(int frameCount, int srcBufPtr, int samplesOfLeadTime)
	{
	int result;
	result = SqSound.InsertSamplesFromLeadTime(frameCount, srcBufPtr, samplesOfLeadTime);
#ifdef SOUND_DEBUG_MSGS
	printf("%d = snd_InsertSamplesFromLeadTime(%d, %d, %d)\n", result, frameCount, srcBufPtr, samplesOfLeadTime);
#endif
	return result;
	}
	
int snd_PlaySamplesFromAtLength(int frameCount, int arrayIndex, int startIndex)
	{
	int result;
	result = SqSound.PlaySamplesFromAtLength(frameCount, arrayIndex, startIndex);
#ifdef SOUND_DEBUG_MSGS
	printf("%d = snd_PlaySamplesFromAtLength(%d, %d, %d)\n", result, frameCount, arrayIndex, startIndex);
#endif
	return result;
	}     
	
int snd_PlaySilence(void)
	{ 
	int result;
	result = SqSound.PlaySilence();
#ifdef SOUND_DEBUG_MSGS
	printf("%d = snd_PlaySilence()\n", result);
#endif
	return result;
	}
     
int snd_Start(int frameCount, int samplesPerSec, int stereo, int semaIndex)
	{
	int result;
	result = SqSound.Start(frameCount, samplesPerSec, stereo, semaIndex);
#ifdef SOUND_DEBUG_MSGS
	printf("%d = snd_Start(%d, %d, %d, %d)\n", result, frameCount, samplesPerSec, stereo, semaIndex);
#endif
	return result;
	}

int snd_Stop(void)
	{
#ifdef SOUND_DEBUG_MSGS
	printf("snd_Stop()\n");
#endif
	SqSound.Stop();
	return 0;
	}

/* sound input */

bool recordingInProgress = false;

int snd_SetRecordLevel(int level)
	{
#ifdef SOUND_DEBUG_MSGS
	printf("snd_SetRecordLevel()\n");
#endif
	/* set the recording level to a value between 0 (minimum gain) and 1000. */
	if (!recordingInProgress || (level < 0) || (level > 1000)) {
		success(false);
		return 0;  /* noop if not recording */
	} 
	
	// not yet implemented
	success(false);
	
    return 0;
	/* don't fail on error; hardware may not support setting the gain */
	}
	
int snd_StartRecording(int desiredSamplesPerSec, int stereo, int semaIndex)
	/* turn on sound recording, trying to use a sampling rate close to
	   the one specified. semaIndex is the index in the exportedObject
	   array of a semaphore to be signalled when input data is available. */
     FAIL(0)
int snd_StopRecording(void)
	/* turn off sound recording */
     FAIL(0)
double snd_GetRecordingSampleRate(void)
	/* return the actual recording rate; fail if not currently recording */
     FAIL(0)
int snd_RecordSamplesIntoAtLength(int buf, int startSliceIndex, int bufferSizeInBytes)
	/* if data is available, copy as many sample slices as possible into the
	   given buffer starting at the given slice index. do not write past the
	   end of the buffer, which is buf + bufferSizeInBytes. return the number
	   of slices (not bytes) copied. a slice is one 16-bit sample in mono
	   or two 16-bit samples in stereo. */
     FAIL(0)
